# coding: utf8

"""
Data handling scripts
"""


def create_merge_file(bids_dir, out_tsv, caps_dir=None, tsv_file=None, pipelines=None, **kwargs):
    """
    Merge all the .TSV files containing clinical data of a BIDS compliant dataset and store
    the result inside a .TSV file.

    Args:
        bids_dir: path to the BIDS folder
        out_tsv: path to the output tsv file
        caps_dir: path to the CAPS folder (optional)
        tsv_file: TSV file containing the subjects with their sessions (optional)
        pipelines: when adding CAPS information, indicates the pipelines that will be merged (optional)

    """
    from os import path
    from glob import glob
    import os
    import pandas as pd
    import numpy as np
    import warnings
    from .pipeline_handling import InitException, DatasetError
    from clinica.utils.participant import get_subject_session_list

    if caps_dir is not None:
        if not path.isdir(caps_dir):
            raise IOError('The path to the CAPS directory is wrong')

    col_list = []
    scans_dict = {}

    if not os.path.isfile(path.join(bids_dir, 'participants.tsv')):
        raise IOError('participants.tsv not found in the specified BIDS directory')
    participants_df = pd.read_csv(path.join(bids_dir, 'participants.tsv'), sep='\t')

    sessions, subjects = get_subject_session_list(bids_dir, ss_file=tsv_file, use_session_tsv=True)
    n_sessions = len(sessions)

    # Find what is dir and what is file_name
    if os.sep not in out_tsv:
        out_dir = os.getcwd()
        if out_tsv == '.':
            out_file_name = 'merge_tsv.tsv'
        else:
            out_file_name = out_tsv
    else:
        out_file_name = out_tsv.split(os.sep)[-1]
        out_dir = path.dirname(out_tsv)

    if len(out_file_name) == 0:
        out_file_name = 'merge_tsv.tsv'

    if '.' not in out_file_name:
        out_file_name = out_file_name + '.tsv'
    else:
        extension = os.path.splitext(out_file_name)[1]
        if extension != '.tsv':
            raise TypeError('Output file must be .tsv.')

    if not os.path.exists(out_dir):
        os.makedirs(out_dir)

    for col in participants_df.columns.values:
        col_list.append(col)

    merged_df = pd.DataFrame(columns=col_list)

    # BIDS part
    i_subject = 0
    while i_subject < n_sessions:
        sub_path = path.join(bids_dir, subjects[i_subject])
        sub_name = sub_path.split(os.sep)[-1]
        # For each subject, extract the relative row from the dataframe
        row_participant = participants_df[participants_df['participant_id'] == sub_name]
        # Open the sessions file related to the subject
        sessions_df = pd.read_csv(path.join(sub_path, sub_name + '_sessions.tsv'), sep='\t')
        # Looking for the sessions corresponding to the subject
        loc_sessions = []
        i_session = i_subject
        while i_session < n_sessions and subjects[i_session] == subjects[i_subject]:
            loc_sessions.append(i_session)
            i_session += 1

        # For each session found
        # extract the information contained in the scans files
        # for line in range(0, len(sessions_df)):
        for i_session in loc_sessions:
            # Extract and convert to a dictionary information
            # regarding the session
            row_session_df = sessions_df[sessions_df.session_id == sessions[i_session]]
            row_session_df.reset_index(inplace=True, drop=True)
            if len(row_session_df) == 0:
                raise DatasetError(sessions_df.loc[0, 'session_id'] + ' / ' + sessions[i_session])

            new_cols = [s for s in row_session_df.columns.values if s not in col_list]
            if len(new_cols) != 0:
                for i in range(0, len(new_cols)):
                    col_list.append(new_cols[i])

            session_id = row_session_df.loc[0, 'session_id']
            if os.path.isfile(path.join(bids_dir, sub_name, 'ses-' + session_id,
                                        sub_name + '_' + 'ses-' + session_id + '_scans.tsv')):
                scans_df = pd.read_csv(path.join(bids_dir, sub_name, 'ses-' + session_id,
                                                 sub_name + '_' + 'ses-' + session_id + '_scans.tsv'), sep='\t')
                for i in range(0, len(scans_df)):
                    for col in scans_df.columns.values:
                        if col == 'filename':
                            pass
                        else:
                            file_scan = scans_df.iloc[i]['filename']
                            file_name = file_scan.split('/')[1]
                            # Remove the extension .nii.gz
                            file_name = os.path.splitext(os.path.splitext(file_name)[0])[0]
                            file_parts = file_name.split('_')
                            last_pattern_index = len(file_parts) - 1
                            mod_type = file_parts[last_pattern_index]
                            value = scans_df.iloc[i][col]
                            new_col_name = col + '_' + mod_type
                            scans_dict.update({new_col_name: value})
                row_scans = pd.DataFrame(scans_dict, index=[0])
            else:
                row_scans = pd.DataFrame()

            new_cols = [s for s in row_scans.columns.values if s not in col_list]
            if len(new_cols) != 0:
                for i in range(0, len(new_cols)):
                    col_list.append(new_cols[i])

            row_to_append_df = pd.DataFrame(columns=row_participant.columns)
            for col in row_participant:
                row_to_append_df[col] = row_participant[col]

            # Append all the data inside session_df
            for col in row_session_df:
                row_to_append_df[col] = row_session_df[col].values[0]

            for col in row_scans:
                row_to_append_df[col] = row_scans[col].values[0]

            merged_df = merged_df.append(row_to_append_df)
        scans_dict = {}
        i_subject = loc_sessions[-1] + 1

    old_index = col_list.index('session_id')
    col_list.insert(1, col_list.pop(old_index))
    merged_df = merged_df[col_list]
    merged_df.to_csv(path.join(out_dir, out_file_name), sep='\t', index=False)

    len_BIDS = len(merged_df.columns)

    # # Call the script for computing the missing modalities
    # # and append the result to the merged file
    # compute_missing_mods(bids_dir, out_dir, 'tmpG7VIY0')
    # tmp_ses = glob(path.join(out_dir, 'tmpG7VIY0*'))
    # for f in tmp_ses:
    #     # Skip the summary file
    #     if 'summary' not in f:
    #         # Load the file
    #         mss_df = pd.read_csv(f, sep='\t')
    #         f_name = f.split(os.sep)[-1]
    #         patterns = f_name.split('-')
    #         ses_id = patterns[len(patterns) - 1]
    #         ses_id = ses_id.replace('.tsv', '')
    #         cols = mss_df.columns.values
    #
    #         # If the file opened contains new columns,
    #         # add them to the existing merged_df
    #         for col_name in cols:
    #             if col_name not in col_list:
    #                 merged_df[col_name] = 0
    #
    #         for i in range(0, len(mss_df)):
    #             row = mss_df.iloc[i]
    #             subj_idx = merged_df[(merged_df['participant_id'] == row['participant_id']) & (
    #                     merged_df['session_id'] == ses_id)].index.tolist()
    #
    #             if len(subj_idx) > 1:
    #                 raise ValueError('Multiple row for the same visit in the merge-tsv file.')
    #             elif len(subj_idx) == 0:
    #                 print 'Warning: Found modalities missing information for the subject:' + row[
    #                     'participant_id'] + ' visit:' + ses_id + ' but the subject is not included in the column participant_id.'
    #                 continue
    #             else:
    #                 subj_idx = subj_idx[0]
    #             for col_name in cols:
    #                 if not col_name == 'participant_id':
    #                     merged_df.iloc[subj_idx, merged_df.columns.get_loc(col_name)] = row[col_name]
    #
    # # Remove all the temporary files created
    # for f in tmp_ses:
    #     os.remove(f)
    #
    # if true_false_mode:
    #     merged_df = merged_df.replace(['Y', 'N'], ['0', '1'])

    merged_df = merged_df.reset_index(drop=True)

    # CAPS
    if caps_dir is not None:
        # Call the different pipelines
        from .pipeline_handling import t1_volume_pipeline, pet_volume_pipeline

        pipeline_options = {
            't1-volume': t1_volume_pipeline,
            'pet-volume': pet_volume_pipeline
        }
        columns_summary = ['pipeline_name', 'group_id', 'atlas_id', 'regions_number', 'first_column_name', 'last_column_name']
        merged_summary_df = pd.DataFrame(columns=columns_summary)
        if pipelines is None:
            for key, pipeline in pipeline_options.items():
                try:
                    merged_df, summary_df = pipeline(caps_dir, merged_df, **kwargs)
                    merged_summary_df = pd.concat([merged_summary_df, summary_df])

                except InitException:
                    warnings.warn('This pipeline was not initialized: ' + key)
        else:
            for pipeline in pipelines:
                merged_df, summary_df = pipeline_options[pipeline](caps_dir, merged_df, **kwargs)
                merged_summary_df = pd.concat([merged_summary_df, summary_df])

        n_atlas = len(merged_summary_df)
        index_column_df = pd.DataFrame(index=np.arange(n_atlas), columns=['first_column_index', 'last_column_index'])
        index_column_df.iat[0, 0] = len_BIDS
        index_column_df.iat[n_atlas - 1, 1] = np.shape(merged_df)[1] - 1
        for i in range(1, n_atlas):
            index_column_df.iat[i, 0] = index_column_df.iat[i-1, 0] + merged_summary_df.iat[i-1, 3]
            index_column_df.iat[i-1, 1] = index_column_df.iat[i, 0] - 1

        merged_summary_df.reset_index(inplace=True, drop=True)
        merged_summary_df = pd.concat([merged_summary_df, index_column_df], axis=1)
        summary_filename = out_file_name.split('.')[0] + '_summary.tsv'
        merged_summary_df.to_csv(path.join(out_dir, summary_filename), sep='\t', index=False)

    merged_df.to_csv(path.join(out_dir, out_file_name), sep='\t', index=False)


def find_mods_and_sess(bids_dir):
    """
    Find all the modalities and sessions available for a given BIDS dataset

    Args:
        bids_dir: path to the BIDS dataset

    Returns:
        mods_dict: a dictionary that stores the sessions and modalities found and has the following structure.
    Example:
    {
        'sessions': ['ses-M00', 'ses-M18'],
        'fmap': ['fmap'],
        'anat': ['flair', 't1w'],
        'func': ['func_task-rest'],
        'dwi': ['dwi']
    }

    """
    from glob import glob
    from os import path
    import os

    mods_dict = {}
    mods_list = []
    subjects_paths_lists = glob(path.join(bids_dir, '*sub-*'))

    for sub_path in subjects_paths_lists:
        ses_paths = glob(path.join(sub_path, '*ses-*'))
        for session in ses_paths:
            ses_name = session.split(os.sep)[-1]
            mods_avail = []
            if 'sessions' in mods_dict:
                if ses_name not in mods_dict['sessions']:
                    mods_dict['sessions'].append(ses_name)
            else:
                mods_dict.update({'sessions': [ses_name]})
            mods_paths_folders = glob(path.join(session, '*/'))

            for p in mods_paths_folders:
                p = p[:-1]
                mods_avail.append(p.split('/').pop())

            if 'func' in mods_avail:
                list_funcs_paths = glob(path.join(session, 'func', '*bold.nii.gz'))
                for func_path in list_funcs_paths:
                    func_name = func_path.split(os.sep)[-1]
                    func_name_tokens = func_name.split('_')
                    func_task = func_name_tokens[2]
                if 'func' in mods_dict:
                    if 'func_' + func_task not in mods_dict['func']:
                        mods_dict['func'].append('func_' + func_task)
                else:
                    mods_dict.update({'func': ['func_' + func_task]})

                if 'func_' + func_task not in mods_list:
                    mods_list.append('func_' + func_task)

            if 'dwi' in mods_avail:
                if 'dwi' not in mods_dict:
                    mods_dict.update({'dwi': ['dwi']})
                if 'dwi' not in mods_list:
                    mods_list.append('dwi')

            if 'fmap' in mods_avail:
                if 'fmap' not in mods_dict:
                    mods_dict.update({'fmap': ['fmap']})
                if 'fmap' not in mods_list:
                    mods_list.append('fmap')

            if 'pet' in mods_avail:
                if 'pet' not in mods_dict:
                    mods_dict.update({'pet': ['pet']})
                if 'pet' not in mods_list:
                    mods_list.append('pet')

            if 'anat' in mods_avail:
                anat_files_paths = glob(path.join(session, 'anat', '*'))

                for anat_file in anat_files_paths:
                    anat_name = anat_file.split(os.sep)[-1]

                    # Extract the name of the file without the extension
                    if '.nii.gz' in anat_name:
                        anat_name = anat_name.replace('.nii.gz', '')
                        anat_ext = 'nii.gz'
                    else:
                        anat_name = os.path.splitext(anat_name.split(os.sep)[-1])[0]
                        anat_ext = os.path.splitext(anat_name.split(os.sep)[-1])[1]

                    if anat_ext != 'json':
                        file_parts = anat_name.split("_")
                        anat_type = str.lower(file_parts[len(file_parts) - 1])
                        if 'anat' in mods_dict:
                            if anat_type not in mods_dict['anat']:
                                anat_aval = mods_dict['anat']
                                anat_aval.append(anat_type)
                                mods_dict.update({'anat': anat_aval})
                        else:
                            mods_dict.update({'anat': [anat_type]})

                        if anat_type not in mods_list:
                            mods_list.append(anat_type)

    return mods_dict


def compute_missing_mods(bids_dir, out_dir, output_prefix=''):
    """
    Compute the list of missing modalities for each subject in a BIDS compliant dataset

    Args:
        bids_dir: path to the BIDS directory
        out_dir: path to the output folder
        output_prefix: string that replace the default prefix ('missing_mods_') in the name of all the output files
    created
    """
    from ..converter_utils import MissingModsTracker, print_statistics
    import os
    from os import path
    import pandas as pd
    from glob import glob

    if not os.path.exists(out_dir):
        os.makedirs(out_dir)

    # Find all the modalities and sessions available for the input dataset
    mods_and_sess = find_mods_and_sess(bids_dir)
    sessions_found = mods_and_sess['sessions']
    mods_and_sess.pop('sessions')
    mods_avail_dict = mods_and_sess
    mods_avail = [j for i in mods_avail_dict.values() for j in i]
    cols_dataframe = mods_avail[:]
    cols_dataframe.insert(0, 'participant_id')
    mmt = MissingModsTracker(sessions_found, mods_avail)

    if output_prefix == '':
        out_file_name = 'missing_mods_'
    else:
        out_file_name = output_prefix + '_'

    summary_file = open(path.join(out_dir, out_file_name + 'summary.txt'), 'w')
    missing_mods_df = pd.DataFrame(columns=cols_dataframe)
    row_to_append_df = pd.DataFrame(columns=cols_dataframe)
    subjects_paths_lists = glob(path.join(bids_dir, '*sub-*'))
    subjects_paths_lists.sort()

    if len(subjects_paths_lists) == 0:
        raise IOError("No subjects found or dataset not BIDS complaint.")
    # Check the modalities available for each session
    for ses in sessions_found:
        mods_avail_bids = []
        for sub_path in subjects_paths_lists:
            mods_avail_bids = []
            subj_id = sub_path.split(os.sep)[-1]
            row_to_append_df['participant_id'] = pd.Series(subj_id)
            ses_path_avail = glob(path.join(sub_path, ses))
            if len(ses_path_avail) == 0:
                mmt.increase_missing_ses(ses)
                for mod in mods_avail:
                    row_to_append_df[mod] = pd.Series('0')
            else:

                ses_path = ses_path_avail[0]
                mods_paths_folders = glob(path.join(ses_path, '*/'))

                for p in mods_paths_folders:
                    p = p[:-1]
                    mods_avail_bids.append(p.split('/').pop())

                # Check if a modality folder is available and if is empty
                if 'func' in mods_avail_bids:
                    # Extract all the task available
                    for m in mods_avail_dict['func']:
                        tokens = m.split('_')
                        task_name = tokens[1]
                        task_avail_list = glob(path.join(
                            ses_path, 'func', '*' + task_name + '*')
                        )

                        if len(task_avail_list) == 0:
                            row_to_append_df[m] = pd.Series('0')
                        else:
                            row_to_append_df[m] = pd.Series('1')
                # If the folder is not available but the modality is
                # in the list of the available one mark it as missing
                else:
                    if 'func' in mods_avail_dict:
                        for m in mods_avail_dict['func']:
                            row_to_append_df[m] = pd.Series('0')
                        mmt.add_missing_mod(ses, m)

                if 'dwi' in mods_avail_bids:
                    row_to_append_df['dwi'] = pd.Series('1')
                else:
                    if 'dwi' in mods_avail:
                        row_to_append_df['dwi'] = pd.Series('0')
                        mmt.add_missing_mod(ses, 'dwi')

                if 'anat' in mods_avail_bids:
                    for m in mods_avail_dict['anat']:
                        anat_aval_list = glob(path.join(ses_path, 'anat', '*.nii.gz'))
                        anat_aval_list = [elem for elem in anat_aval_list if m.lower() in elem.lower()]
                        if len(anat_aval_list) > 0:
                            row_to_append_df[m] = pd.Series('1')
                        else:
                            row_to_append_df[m] = pd.Series('0')
                            mmt.add_missing_mod(ses, m)
                else:
                    if 'anat' in mods_avail_dict:
                        for m in mods_avail_dict['anat']:
                            row_to_append_df[m] = pd.Series('0')
                            mmt.add_missing_mod(ses, m)

                if 'fmap' in mods_avail_bids:
                    row_to_append_df['fmap'] = pd.Series('1')
                else:
                    if 'fmap' in mods_avail:
                        row_to_append_df['fmap'] = pd.Series('0')
                        mmt.add_missing_mod(ses, 'fmap')
                if 'pet' in mods_avail_bids:
                    row_to_append_df['pet'] = pd.Series('1')
                else:
                    if 'pet' in mods_avail:
                        row_to_append_df['pet'] = pd.Series('0')
                        mmt.add_missing_mod(ses, 'pet')

            missing_mods_df = missing_mods_df.append(row_to_append_df)
            row_to_append_df = pd.DataFrame(columns=cols_dataframe)

        missing_mods_df = missing_mods_df[cols_dataframe]
        missing_mods_df.to_csv(path.join(out_dir, out_file_name + ses + '.tsv'), sep='\t', index=False,
                               encoding='utf-8')
        missing_mods_df = pd.DataFrame(columns=cols_dataframe)

    print_statistics(summary_file, len(subjects_paths_lists), sessions_found, mmt)


def create_subs_sess_list(input_dir, output_dir,
                          file_name=None, is_bids_dir=True, use_session_tsv=False):
    """
    Create the file subject_session_list.tsv that contains the list
    of the visits for each subject for a BIDS or CAPS compliant dataset.

    Args:
        input_dir (str): Path to the BIDS or CAPS directory.
        output_dir (str): Path to the output directory
        file_name: name of the output file
        is_bids_dir (boolean): Specify if input_dir is a BIDS directory or
            not (i.e. a CAPS directory)
        use_session_tsv (boolean): Specify if the list uses the sessions listed in the sessions.tsv files
    """
    from os import path
    from glob import glob
    import pandas as pd
    import os

    if not os.path.exists(output_dir):
        os.makedirs(output_dir)

    if file_name is None:
        file_name = 'subjects_sessions_list.tsv'
    subjs_sess_tsv = open(path.join(output_dir, file_name), 'w')
    subjs_sess_tsv.write('participant_id' + '\t' + 'session_id' + '\n')

    if is_bids_dir:
        path_to_search = input_dir
    else:
        path_to_search = path.join(input_dir, 'subjects')
    subjects_paths = glob(path.join(path_to_search, '*sub-*'))

    # Sort the subjects list
    subjects_paths.sort()

    if len(subjects_paths) == 0:
        raise IOError('Dataset empty or not BIDS/CAPS compliant.')

    for sub_path in subjects_paths:
        subj_id = sub_path.split(os.sep)[-1]

        if use_session_tsv:
            session_df = pd.read_csv(path.join(sub_path, subj_id + '_sessions.tsv'), sep='\t')
            session_list = list(session_df['session_id'].to_numpy())
            for session in session_list:
                subjs_sess_tsv.write(subj_id + '\t' + session + '\n')

        else:
            sess_list = glob(path.join(sub_path, '*ses-*'))

            for ses_path in sess_list:
                session_name = ses_path.split(os.sep)[-1]
                subjs_sess_tsv.write(subj_id + '\t' + session_name + '\n')

    subjs_sess_tsv.close()


def center_nifti_origin(input_image, output_image):
    """

    Put the origin of the coordinate system at the center of the image

    Args:
        input_image: path to the input image
        output_image: path to the output image (where the result will be stored)

    Returns:
        path of the output image created
    """

    import nibabel as nib
    import numpy as np
    from colorama import Fore
    from nibabel.spatialimages import ImageFileError
    from os.path import isfile
    import os

    error_str = None
    try:
        img = nib.load(input_image)
    except FileNotFoundError:
        error_str = Fore.RED + '[Error] No such file ' + input_image + Fore.RESET
    except ImageFileError:
        error_str = Fore.RED + '[Error] File ' + input_image + ' could not be read.' + Fore.RESET
    except Exception as e:
        error_str = Fore.RED + '[Error] File ' + input_image + ' could not be loaded with nibabel: ' + \
                    str(e) + Fore.RESET

    if not error_str:
        try:
            canonical_img = nib.as_closest_canonical(img)
            hd = canonical_img.header

            qform = np.zeros((4, 4))
            for i in range(1, 4):
                qform[i - 1, i - 1] = hd['pixdim'][i]
                qform[i - 1, 3] = -1.0 * hd['pixdim'][i] * hd['dim'][i] / 2.0
            new_img = nib.Nifti1Image(canonical_img.get_data(caching='unchanged'), affine=qform, header=hd)

            # Without deleting already-existing file, nib.save causes a severe bug on Linux system
            if isfile(output_image):
                os.remove(output_image)

            nib.save(new_img, output_image)
            if not isfile(output_image):
                error_str = Fore.RED + '[Error] NIfTI file created but Clinica could not save it to ' \
                            + output_image + '. Please check that the output folder has the correct permissions.' \
                            + Fore.RESET
        except Exception as e:
            error_str = Fore.RED + '[Error] File ' + input_image + ' could not be processed with nibabel: ' + \
                        str(e) + Fore.RESET

    return output_image, error_str


def center_all_nifti(bids_dir, output_dir, modality, center_all_files=False):
    """
    Center all the NIfTI images of the input BIDS folder into the empty output_dir specified in argument.
    All the files from bids_dir are copied into output_dir, then all the NIfTI images we can found are replaced by their
    centered version if their center if off the origin by more than 50 mm.

    Args:
        bids_dir: (str) path to bids directory
        output_dir: (str) path to EMPTY output directory
        modality: (list of str) modalities to convert
        center_all_files: (bool) center only files that may cause problem for SPM if false. If true, center all NIfTI

    Returns:
        List of the centered files
    """
    from colorama import Fore
    from clinica.utils.inputs import check_bids_folder
    from clinica.utils.exceptions import ClinicaBIDSError
    from os.path import join, basename
    from glob import glob
    from os import listdir
    from os.path import isdir, isfile
    from shutil import copy2, copytree

    # output and input must be different, so that we do not mess with user's data
    if bids_dir == output_dir:
        raise ClinicaBIDSError(Fore.RED + '[Error] Input BIDS and output directories must be different' + Fore.RESET)

    assert isinstance(modality, list), 'modality arg must be a list of str'

    # check that input is a BIDS dir
    check_bids_folder(bids_dir)

    for f in listdir(bids_dir):
        if isdir(join(bids_dir, f)) and not isdir(join(output_dir, f)):
            copytree(join(bids_dir, f), join(output_dir, f))
        elif isfile(join(bids_dir, f)) and not isfile(join(output_dir, f)):
            copy2(join(bids_dir, f), output_dir)

    pattern = join(output_dir, '**/*.nii*')
    nifti_files = glob(pattern, recursive=True)

    # Now filter this list by elements in modality list
    #   For each file:
    #       if any modality name (lowercase) is found in the basename of the file:
    #           keep the file
    nifti_files_filtered = [f for f in nifti_files
                            if any(elem.lower() in basename(f).lower() for elem in modality)]

    # Remove those who are centered
    if not center_all_files:
        nifti_files_filtered = [file for file in nifti_files_filtered if not is_centered(file)]

    all_errors = []
    for f in nifti_files_filtered:
        print('Handling ' + f)
        _, current_error = center_nifti_origin(f, f)
        if current_error:
            all_errors.append(current_error)
    if len(all_errors) > 0:
        final_error_msg = Fore.RED + '[Error] Clinica encoutered ' + str(len(all_errors)) \
                          + ' error(s) while trying to center all NIfTI images.\n'
        for error in all_errors:
            final_error_msg += '\n' + error
        raise RuntimeError(final_error_msg)
    return nifti_files_filtered


def are_far_appart(file1, file2, threshold=80):
    """
    Tells if 2 files have a center located at more than a threshold distance
    Args:
        file1: (str) path to the first nifti file
        file2: (str) path to the second nifti file
        threshold: threshold to consider whether 2 files are too far appart

    Returns:
        True if distance between `file1` and `file2` is greter than `threshold`, False otherwise.
    """
    from os.path import isfile
    import numpy as np

    assert isfile(file1)
    assert isfile(file2)

    center1 = get_world_coordinate_of_center(file1)
    center2 = get_world_coordinate_of_center(file2)

    return np.linalg.norm(center2 - center1, ord=2) > threshold


def write_list_of_files(file_list, output_file):
    """
    Save `file_list` list of files into `output_file` text file.
    Args:
        file_list: (list of str) of path to files
        output_file: (str) path to the output txt file

    Returns:
        output_file
    """
    from os.path import isfile

    assert isinstance(file_list, list), 'First argument must be a list'
    assert isinstance(output_file, str), 'Second argument must be a str'
    if isfile(output_file):
        return None

    text_file = open(output_file, 'w+')
    for created_file in file_list:
        text_file.write(created_file + '\n')
    text_file.close()
    return output_file


def check_relative_volume_location_in_world_coordinate_system(label_1, nifti_list1,
                                                              label_2, nifti_list2,
                                                              bids_dir,
                                                              modality):
    """
    Check if the NIfTI file list nifti_list1 and nifti_list2 provided in argument are not too far apart (otherwise coreg
    in SPM may fail. Norm between center of volumes of 2 files must be less than 80 mm.

    Args:
        label_1: label of the first nifti_list1 files (used in potential warning message)
        nifti_list1: first set of files
        label_2: label of the second nifti_list
        nifti_list2: second set of files, must be same length as nifti_list1
        bids_dir: bids directory (used in potential warning message)
        modality: string that must be used in argument of: clinica iotools bids --modality MODALITY (used in potential
                warning message)
    Returns:
        Nothing
    """
    import numpy as np
    from colorama import Fore
    from os.path import abspath, basename
    from clinica.utils.stream import cprint
    import sys

    center_coordinate_1 = [get_world_coordinate_of_center(file) for file in nifti_list1]
    center_coordinate_2 = [get_world_coordinate_of_center(file) for file in nifti_list2]

    l2_norm = [np.linalg.norm(center_1 - center_2) for center_1, center_2 in zip(center_coordinate_1, center_coordinate_2)]
    pairs_with_problems = [i for i, norm in enumerate(l2_norm) if norm > 80]

    if len(pairs_with_problems) > 0:
        warning_message = (Fore.YELLOW + '[Warning] It appears that ' + str(len(pairs_with_problems)) + ' pairs of files'
                           + ' have an important relative offset. SPM coregistration has a high probability to fail '
                           + 'on these files:\n\n')

        # File column width : 3 spaces more than the longest string to display
        file1_width = max(3 + len(label_1),
                          3 + max(len(basename(file)) for file in [nifti_list1[k] for k in pairs_with_problems]))
        file2_width = max(3 + len(label_2),
                          3 + max(len(basename(file)) for file in [nifti_list2[k] for k in pairs_with_problems]))

        norm_width = len('Relative distance')

        warning_message += ('%-' + str(file1_width)
                            + 's%-' + str(file2_width)
                            + 's%-' + str(norm_width) + 's') % (label_1,
                                                                label_2,
                                                                'Relative distance')

        warning_message += '\n' + '-' * (file1_width + file2_width + norm_width) + '\n'
        for file1, file2, norm in zip([nifti_list1[k] for k in pairs_with_problems],
                                      [nifti_list2[k] for k in pairs_with_problems],
                                      [l2_norm[k] for k in pairs_with_problems]):
            # Nice formatting as array
            # % escape character
            # - aligned to the left, with the size of the column
            # s = string, f = float
            # . for precision with float
            # https://docs.python.org/2/library/stdtypes.html#string-formatting for more information
            warning_message += ('%-' + str(file1_width)
                                + 's%-' + str(file2_width)
                                + 's%-' + str(norm_width) + '.2f\n') % (str(basename(file1)),
                                                                        str(basename(file2)),
                                                                        norm)
        warning_message += '\nClinica provides a tool to counter this problem by replacing the center of the volume' \
                           + ' at the origin of the world coordinates.\nUse the following command line to correct the '\
                           + 'header of the faulty NIFTI volumes in a new folder:\n' + Fore.RESET \
                           + Fore.BLUE + '\nclinica iotools center-nifti ' + abspath(bids_dir) + ' ' \
                           + abspath(bids_dir) + '_centered --modality "' + modality + '"\n\n'  \
                           + Fore.YELLOW + 'You will find more information on the command by typing ' + Fore.BLUE \
                           + 'clinica iotools center-nifti' + Fore.YELLOW + ' in the console.\nDo you still want to ' \
                           + 'launch the pipeline now?' + Fore.RESET
        cprint(warning_message)
        while True:
            cprint('Your answer [yes/no]:')
            answer = input()
            if answer.lower() in ['yes', 'no']:
                break
            else:
                cprint(Fore.RED + 'You must answer yes or no' + Fore.RESET)
        if answer.lower() == 'no':
            cprint(Fore.RED + 'Clinica will now exit...' + Fore.RESET)
            sys.exit(0)


def check_volume_location_in_world_coordinate_system(nifti_list, bids_dir, modality='t1w'):
    """
    Check if the NIfTI file list nifti_list provided in argument are aproximately centered around the origin of the
    world coordinates. (Problem may arise with SPM segmentation

    If yes, we warn the user of this problem, and propose him to exit clinica in order for him to run:
        clinica iotools center-nifti ...
    or to continue with the execution of the pipeline

    Args:
        nifti_list: (list of str) list of path to nifti files
        bids_dir: (str) path to bids directory associated with this check (in order to propose directly the good
            command line for center-nifti tool)
        modality: (str) to propose directly the good command line option

    Returns:
        Nothing
    """
    from colorama import Fore
    from clinica.utils.stream import cprint
    from os.path import abspath, basename
    import numpy as np
    import sys

    list_non_centered_files = [file for file in nifti_list if not is_centered(file)]
    if len(list_non_centered_files) > 0:
        centers = [get_world_coordinate_of_center(file) for file in list_non_centered_files]
        l2_norm = [np.linalg.norm(center, ord=2) for center in centers]

        # File column width : 3 spaces more than the longest string to display
        file_width = 3 + max(len(basename(file)) for file in list_non_centered_files)
        # Center column width (with a fixed minimum size) : 3 spaces more than the longest string to display
        center_width = max(len('Coordinate of center') + 3,
                           3 + max(len(str(center)) for center in centers))

        warning_message = (Fore.YELLOW + '[Warning] It appears that ' + str(len(list_non_centered_files)) + ' files '
                           + 'have a center way out of the origin of the world coordinate system. SPM has a high prob'
                           + 'ability to fail on these files (for coregistration or segmentation):\n\n')
        warning_message += ('%-' + str(file_width) + 's%-' + str(center_width) + 's%-s') % ('File',
                                                                                            'Coordinate of center',
                                                                                            'Distance to origin')
        # 18 is the length of the string 'Distance to origin'
        warning_message += '\n' + '-' * (file_width + center_width + 18) + '\n'
        for file, center, l2 in zip(list_non_centered_files, centers, l2_norm):
            # Nice formatting as array
            # % escape character
            # - aligned to the left, with the size of the column
            # s = string, f = float
            # . for precision with float
            # https://docs.python.org/2/library/stdtypes.html#string-formatting for more information
            warning_message += ('%-' + str(file_width) + 's%-' + str(center_width) + 's%-25.2f\n') % (basename(file),
                                                                                                      str(center),
                                                                                                      l2)

        cmd_line = (Fore.BLUE
                    + '\nclinica iotools center-nifti '
                    + abspath(bids_dir) + ' '
                    + abspath(bids_dir) + '_centered'
                    + '--modality "' + modality + '"'
                    + '\n\n' + Fore.YELLOW)

        warning_message += ('\nIf you are trying to launch the t1-freesurfer pipeline, you can ignore this message '
                            + 'if you do not want to run the pet-surface pipeline afterward.')

        warning_message += '\nClinica provides a tool to counter this problem by replacing the center of the volume' \
                           + ' at the origin of the world coordinates.\nUse the following command line to correct the '\
                           + 'header of the faulty NIFTI volumes in a new folder:\n'\
                           + cmd_line \
                           + 'You will find more information on the command by typing ' + Fore.BLUE \
                           + 'clinica iotools center-nifti' + Fore.YELLOW + ' in the console.\nDo you still want to '\
                           + 'launch the pipeline now?' + Fore.RESET
        cprint(warning_message)
        while True:
            cprint('Your answer [yes/no]:')
            answer = input()
            if answer.lower() in ['yes', 'no']:
                break
            else:
                cprint(Fore.RED + 'You must answer yes or no' + Fore.RESET)
        if answer.lower() == 'no':
            cprint(Fore.RED + 'Clinica will now exit...' + Fore.RESET)
            sys.exit(0)


def is_centered(nii_volume, threshold_l2=50):
    """
    Tells if a NIfTI volume is centered on the origin of the world coordinate system.

    SPM has troubles to segment files if the center of the volume is not close from the origin of the world coordinate
    system. A series of experiment have been conducted: we take a volume whose center is on the origin of the world
    coordinate system. We add an offset using coordinates of affine matrix [0, 3], [1, 3], [2, 3] (or by modifying the
    header['srow_x'][3], header['srow_y'][3], header['srow_z'][3], this is strictly equivalent).

    It has been determined that volumes were still segmented with SPM when the L2 distance between origin and center of
    the volume did not exceed 100 mm. Above this distance, either the volume is either not segmented (SPM error), or the
    produced segmentation is wrong (not the shape of a brain anymore)

    Args:
        nii_volume: path to NIfTI volume
        threshold_l2: maximum distance between origin of the world coordinate system and the center of the volume to
                    be considered centered. The threshold were SPM segmentation stops working is around 100 mm
                    (it was determined empirically after several trials on a genrated dataset), so default value is 50
                    mm in order to have a security margin, even when dealing with coregistred files afterward)

    Returns:
        True or False

    """
    import numpy as np
    from os.path import basename

    center = get_world_coordinate_of_center(nii_volume)

    # Compare to the threshold and retun boolean
    # if center is a np.nan, comparison will be False, and False will be returned
    distance_from_origin = np.linalg.norm(center, ord=2)
    # if not np.isnan(distance_from_origin):
    #     print('\t' + basename(nii_volume) + ' has its center at {0:.2f} mm of the origin.'.format(distance_from_origin))
    if distance_from_origin < threshold_l2:
        return True
    else:
        # If center is a np.nan,
        return False


def get_world_coordinate_of_center(nii_volume):
    """
    Extract the world coordinates of the center of the image. Based on methods described
    here : https://brainder.org/2012/09/23/the-nifti-file-format/

    Args:
        nii_volume: path to nii volume

    Returns:

    """
    from os.path import isfile
    import nibabel as nib
    from colorama import Fore
    import numpy as np

    assert isinstance(nii_volume, str), 'input argument nii_volume must be a str'
    assert isfile(nii_volume), 'input argument must be a path to a file'

    try:
        orig_nifti = nib.load(nii_volume)
    except nib.filebasedimages.ImageFileError:
        print(Fore.RED + '[Error] ' + nii_volume
              + ' could not be read by nibabel. Is it a valid NIfTI file ?' + Fore.RESET)
        return np.nan

    head = orig_nifti.header

    if isinstance(head, nib.freesurfer.mghformat.MGHHeader):
        # If MGH volume
        center_coordinates_world = vox_to_world_space_method_3_bis(head['dims'][0:3] / 2, head)
    else:
        # Standard NIfTI volume
        center_coordinates = get_center_volume(head)

        if head['qform_code'] > 0:
            center_coordinates_world = vox_to_world_space_method_2(center_coordinates, head)
        elif head['sform_code'] > 0:
            center_coordinates_world = vox_to_world_space_method_3(center_coordinates, head)
        elif head['sform_code'] == 0:
            center_coordinates_world = vox_to_world_space_method_1(center_coordinates, head)
        else:
            center_coordinates_world = np.nan
    return center_coordinates_world


def get_center_volume(header):
    """
    Get the voxel coordinates of the center of the data, using header information
    Args:
        header: a nifti header

    Returns:
        Voxel coordinates of the center of the volume
    """
    import numpy as np

    center_x = header['dim'][1] / 2
    center_y = header['dim'][2] / 2
    center_z = header['dim'][3] / 2
    return np.array([center_x,
                     center_y,
                     center_z])


def vox_to_world_space_method_1(coordinates_vol, header):
    """
    The Method 1 is for compatibility with analyze and is not supposed to be used as the main orientation method. But it
    is used if sform_code = 0. The world coordinates are determined simply by scaling by the voxel size by their
    dimension stored in pixdim. More information here: https://brainder.org/2012/09/23/the-nifti-file-format/
    Args:
        coordinates_vol: coordinate in the volume (raw data)
        header: header object

    Returns:
        Coordinates in the world space
    """
    import numpy as np

    return np.array(coordinates_vol) * np.array(header['pixdim'][1],
                                                header['pixdim'][2],
                                                header['pixdim'][3])


def vox_to_world_space_method_2(coordinates_vol, header):
    """
    The Method 2 is used when short qform_code is larger than zero. To get the coordinates, we multiply a rotation
    matrix (r_mat) by coordinates_vol, then perform hadamart with pixel dimension pixdim (like in method 1). Then we add
    an offset (qoffset_x, qoffset_y, qoffset_z)

    Args:
        coordinates_vol: coordinate in the volume (raw data)
        header: header object

    Returns:
        Coordinates in the world space
    """
    import numpy as np

    def get_r_matrix(h):
        """
        Get rotation matrix, more information here: https://brainder.org/2012/09/23/the-nifti-file-format/
        Args:
            h: header

        Returns:
            Rotation matrix
        """
        b = h['quatern_b']
        c = h['quatern_c']
        d = h['quatern_d']
        a = np.sqrt(1 - (b ** 2) - (c ** 2) - (d ** 2))
        r = np.zeros((3, 3))
        r[0, 0] = (a ** 2) + (b ** 2) - (c ** 2) - (d ** 2)
        r[0, 1] = 2 * ((b * c) - (a * d))
        r[0, 2] = 2 * ((b * d) + (a * c))
        r[1, 0] = 2 * ((b * c) + (a * d))
        r[1, 1] = (a ** 2) + (c ** 2) - (b ** 2) - (d ** 2)
        r[1, 2] = 2 * ((c * d) - (a * b))
        r[2, 0] = 2 * ((b * d) - (a * c))
        r[2, 1] = 2 * ((b * d) - (a * c))
        r[2, 2] = (a ** 2) + (d ** 2) - (b ** 2) - (c ** 2)
        return r
    i = coordinates_vol[0]
    j = coordinates_vol[1]
    k = coordinates_vol[2]
    if header['qform_code'] > 0:
        r_mat = get_r_matrix(header)
    else:
        # Should never be reached
        raise ValueError('qform_code must be greater than 0 to use this method')
    q = header['pixdim'][0]
    if q not in [-1, 1]:
        print('q was ' + str(q), ', now is 1')
        q = 1
    return np.dot(r_mat, np.array([i, j, q * k])) * np.array(header['pixdim'][1:4]) + np.array([header['qoffset_x'],
                                                                                                header['qoffset_y'],
                                                                                                header['qoffset_z']])


def vox_to_world_space_method_3(coordinates_vol, header):
    """
    This method is used when sform_code is larger than zero. It relies on a full affine matrix, stored in the header in
     the fields srow_[x,y,y], to map voxel to world coordinates.
     When a nifti file is created with raw data and affine=..., this is this method that is used to decypher the
     voxel-to-world correspondance.
    Args:
        coordinates_vol: coordinate in the volume (raw data)
        header: header object

    Returns:
        Coordinates in the world space

    """
    import numpy as np

    def get_aff_matrix(h):
        """
        Get affine transformation matrix, described here : https://brainder.org/2012/09/23/the-nifti-file-format/
        Args:
            h: header

        Returns:
            affine transformation matrix
        """
        mat = np.zeros((4, 4))
        mat[0, 0] = h['srow_x'][0]
        mat[0, 1] = h['srow_x'][1]
        mat[0, 2] = h['srow_x'][2]
        mat[0, 3] = h['srow_x'][3]
        mat[1, 0] = h['srow_y'][0]
        mat[1, 1] = h['srow_y'][1]
        mat[1, 2] = h['srow_y'][2]
        mat[1, 3] = h['srow_y'][3]
        mat[2, 0] = h['srow_z'][0]
        mat[2, 1] = h['srow_z'][1]
        mat[2, 2] = h['srow_z'][2]
        mat[2, 3] = h['srow_z'][3]
        mat[3, 3] = 1
        return mat

    if header['sform_code'] > 0:
        aff = get_aff_matrix(header)
    else:
        # Should never be reached
        raise ValueError('sform_code has a value > 0, so method 3 cannot be used')

    homogeneous_coord = np.concatenate((np.array(coordinates_vol), np.array([1])), axis=0)
    return np.dot(aff, homogeneous_coord)[0:3]


def vox_to_world_space_method_3_bis(coordinates_vol, header):
    """
    This method relies on the same technique as method 3, but for images created by FreeSurfer (MGHImage, MGHHeader)
    Args:
        coordinates_vol: coordinate in the volume (raw data)
        header: nib.freesurfer.mghformat.MGHHeader object

    Returns:
        Coordinates in the world space
    """
    import numpy as np

    affine_trensformation_matrix = header.get_affine()
    homogeneous_coord = np.concatenate((np.array(coordinates_vol), np.array([1])), axis=0)
    return np.dot(affine_trensformation_matrix, homogeneous_coord)[0:3]



# coding: utf8

import clinica.pipelines.engine as cpe

# Use hash instead of parameters for iterables folder names
# Otherwise path will be too long and generate OSError
from nipype import config
cfg = dict(execution={'parameterize_dirs': False})
config.update_config(cfg)


class T1VolumeTissueSegmentation(cpe.Pipeline):
    """T1VolumeTissueSegmentation - Tissue segmentation, bias correction and
    spatial normalization to MNI space.

    Returns:
        A clinica pipeline object containing the T1VolumeTissueSegmentation pipeline.
    """
    def check_custom_dependencies(self):
        """Check dependencies that can not be listed in the `info.json` file."""
        pass

    def check_pipeline_parameters(self):
        """Check pipeline parameters."""
        from clinica.utils.spm import get_tpm

        if 'tissue_classes' not in self.parameters.keys():
            self.parameters['tissue_classes'] = [1, 2, 3]
        if 'dartel_tissues' not in self.parameters.keys():
            self.parameters['dartel_tissues'] = [1, 2, 3]
        if 'tissue_probability_maps' not in self.parameters.keys():
            self.parameters['tissue_probability_maps'] = None
        if 'save_warped_unmodulated' not in self.parameters.keys():
            self.parameters['save_warped_unmodulated'] = True
        if 'save_warped_modulated' not in self.parameters.keys():
            self.parameters['save_warped_modulated'] = False

        # Get Tissue Probability Map from SPM
        if self.parameters['tissue_probability_maps'] is None:
            self.parameters['tissue_probability_maps'] = get_tpm()

    def get_input_fields(self):
        """Specify the list of possible inputs of this pipelines.

        Returns:
            A list of (string) input fields name.
        """
        return ['t1w']

    def get_output_fields(self):
        """Specify the list of possible outputs of this pipelines.

        Returns:
            A list of (string) output fields name.
        """
        return ['bias_corrected_images',
                'bias_field_images',
                'dartel_input_images',
                'forward_deformation_field',
                'inverse_deformation_field',
                'modulated_class_images',
                'native_class_images',
                'normalized_class_images',
                'transformation_mat',
                't1_mni']

    def build_input_node(self):
        """Build and connect an input node to the pipeline.

        Raise:
            ClinicaBIDSError: If there are duplicated files or missing files for any subject
        """
        import nipype.interfaces.utility as nutil
        import nipype.pipeline.engine as npe
        from clinica.utils.exceptions import ClinicaBIDSError, ClinicaException
        from clinica.utils.stream import cprint
        from clinica.iotools.utils.data_handling import check_volume_location_in_world_coordinate_system
        from clinica.utils.inputs import clinica_file_reader
        from clinica.utils.input_files import T1W_NII
        from clinica.utils.ux import print_images_to_process

        # Inputs from anat/ folder
        # ========================
        # T1w file:
        try:
            t1w_files = clinica_file_reader(self.subjects,
                                            self.sessions,
                                            self.bids_directory,
                                            T1W_NII)
        except ClinicaException as e:
            err = 'Clinica faced error(s) while trying to read files in your BIDS directory.\n' + str(e)
            raise ClinicaBIDSError(err)

        check_volume_location_in_world_coordinate_system(t1w_files, self.bids_directory)

        if len(self.subjects):
            print_images_to_process(self.subjects, self.sessions)
            cprint('The pipeline will last approximately 10 minutes per image.')

        read_node = npe.Node(name="ReadingFiles",
                             iterables=[
                                 ('t1w', t1w_files),
                             ],
                             synchronize=True,
                             interface=nutil.IdentityInterface(
                                 fields=self.get_input_fields())
                             )
        self.connect([
            (read_node, self.input_node, [('t1w', 't1w')]),
        ])

    def build_output_node(self):
        """Build and connect an output node to the pipeline."""
        pass

    def build_core_nodes(self):
        """Build and connect the core nodes of the pipeline."""
        import nipype.pipeline.engine as npe
        import nipype.interfaces.utility as nutil
        import nipype.interfaces.io as nio
        import nipype.interfaces.spm as spm
        from ..t1_volume_tissue_segmentation import t1_volume_tissue_segmentation_utils as seg_utils
        from clinica.utils.filemanip import unzip_nii, zip_nii
        from clinica.utils.nipype import fix_join
        from clinica.utils.spm import spm_standalone_is_available, use_spm_standalone

        if spm_standalone_is_available():
            use_spm_standalone()

        # Get <subject_id> (e.g. sub-CLNC01_ses-M00) from input_node
        # and print begin message
        # =======================
        init_node = npe.Node(interface=nutil.Function(
            input_names=self.get_input_fields(),
            output_names=['subject_id'] + self.get_input_fields(),
            function=seg_utils.init_input_node),
            name='0-InitNode')

        # Unzipping
        # =========
        unzip_node = npe.Node(nutil.Function(input_names=['in_file'],
                                             output_names=['out_file'],
                                             function=unzip_nii),
                              name='1-UnzipT1w')

        # Unified Segmentation
        # ====================
        new_segment = npe.Node(spm.NewSegment(),
                               name='2-SpmSegmentation')
        new_segment.inputs.write_deformation_fields = [True, True]
        new_segment.inputs.tissues = seg_utils.get_tissue_tuples(
            self.parameters['tissue_probability_maps'],
            self.parameters['tissue_classes'],
            self.parameters['dartel_tissues'],
            self.parameters['save_warped_unmodulated'],
            self.parameters['save_warped_modulated']
        )

        # Apply segmentation deformation to T1 (into MNI space)
        # =====================================================
        t1_to_mni = npe.Node(seg_utils.ApplySegmentationDeformation(),
                             name='3-T1wToMni')

        # Print end message
        # =================
        print_end_message = npe.Node(
            interface=nutil.Function(
                input_names=['subject_id', 'final_file'],
                function=seg_utils.print_end_pipeline),
            name='WriteEndMessage')

        # Connection
        # ==========
        self.connect([
            (self.input_node, init_node, [('t1w', 't1w')]),
            (init_node, unzip_node, [('t1w', 'in_file')]),
            (unzip_node, new_segment, [('out_file', 'channel_files')]),
            (init_node, print_end_message, [('subject_id', 'subject_id')]),
            (unzip_node, t1_to_mni, [('out_file', 'in_files')]),
            (new_segment, t1_to_mni, [('forward_deformation_field', 'deformation_field')]),
            (new_segment, self.output_node, [('bias_corrected_images', 'bias_corrected_images'),
                                             ('bias_field_images', 'bias_field_images'),
                                             ('dartel_input_images', 'dartel_input_images'),
                                             ('forward_deformation_field', 'forward_deformation_field'),
                                             ('inverse_deformation_field', 'inverse_deformation_field'),
                                             ('modulated_class_images', 'modulated_class_images'),
                                             ('native_class_images', 'native_class_images'),
                                             ('normalized_class_images', 'normalized_class_images'),
                                             ('transformation_mat', 'transformation_mat')]),
            (t1_to_mni, self.output_node, [('out_files', 't1_mni')]),
            (self.output_node, print_end_message, [('t1_mni', 'final_file')]),
        ])

        # Find container path from t1w filename
        # =====================================
        container_path = npe.Node(
            nutil.Function(input_names=['t1w_filename'],
                           output_names=['container'],
                           function=seg_utils.t1w_container_from_filename),
            name='ContainerPath')

        # Writing CAPS
        # ============
        write_node = npe.Node(name='WriteCAPS',
                              interface=nio.DataSink()
                              )
        write_node.inputs.base_directory = self.caps_directory
        write_node.inputs.parameterization = False
        write_node.inputs.regexp_substitutions = [
            (r'(.*)c1(sub-.*)(\.nii(\.gz)?)$',
             r'\1\2_segm-graymatter\3'),
            (r'(.*)c2(sub-.*)(\.nii(\.gz)?)$',
             r'\1\2_segm-whitematter\3'),
            (r'(.*)c3(sub-.*)(\.nii(\.gz)?)$',
             r'\1\2_segm-csf\3'),
            (r'(.*)c4(sub-.*)(\.nii(\.gz)?)$',
             r'\1\2_segm-bone\3'),
            (r'(.*)c5(sub-.*)(\.nii(\.gz)?)$',
             r'\1\2_segm-softtissue\3'),
            (r'(.*)c6(sub-.*)(\.nii(\.gz)?)$',
             r'\1\2_segm-background\3'),
            (r'(.*)(/native_space/sub-.*)(\.nii(\.gz)?)$',
             r'\1\2_probability\3'),
            (r'(.*)(/([a-z]+)_deformation_field/)i?y_(sub-.*)(\.nii(\.gz)?)$',
             r'\1/normalized_space/\4_target-Ixi549Space_transformation-\3_deformation\5'),
            (r'(.*)(/t1_mni/)w(sub-.*)_T1w(\.nii(\.gz)?)$', r'\1/normalized_space/\3_space-Ixi549Space_T1w\4'),
            (r'(.*)(/modulated_normalized/)mw(sub-.*)(\.nii(\.gz)?)$',
             r'\1/normalized_space/\3_space-Ixi549Space_modulated-on_probability\4'),
            (r'(.*)(/normalized/)w(sub-.*)(\.nii(\.gz)?)$',
             r'\1/normalized_space/\3_space-Ixi549Space_modulated-off_probability\4'),
            (r'(.*/dartel_input/)r(sub-.*)(\.nii(\.gz)?)$',
             r'\1\2_dartelinput\3'),
            # Will remove trait_added empty folder
            (r'trait_added', r'')
        ]

        self.connect([
            (self.input_node, container_path, [('t1w', 't1w_filename')]),
            (container_path, write_node, [(('container', fix_join, ''), 'container')]),
            (self.output_node, write_node, [(('native_class_images', seg_utils.zip_list_files, True), 'native_space'),
                                            (('dartel_input_images', seg_utils.zip_list_files, True), 'dartel_input')]),
            (self.output_node, write_node, [
                (('inverse_deformation_field', zip_nii, True), 'inverse_deformation_field')
            ]),
            (self.output_node, write_node, [
                (('forward_deformation_field', zip_nii, True), 'forward_deformation_field')
            ]),
            (self.output_node, write_node, [(('t1_mni', zip_nii, True), 't1_mni')]),
        ])
        if self.parameters['save_warped_unmodulated']:
            self.connect([
                (self.output_node, write_node, [
                    (('normalized_class_images', seg_utils.zip_list_files, True), 'normalized')
                ]),
            ])
        if self.parameters['save_warped_modulated']:
            self.connect([
                (self.output_node, write_node, [
                    (('modulated_class_images', seg_utils.zip_list_files, True), 'modulated_normalized')
                ]),
            ])



# coding: utf8


from clinica.iotools.converters.nifd_to_bids.utils.descriptor import Descriptor


class Old_image_folder(object):
    def __init__(self, strg):

        self.name = strg
        self.new_name = False
        self.old_name = None

        self.bids_info = None

    def get_new_name(self, desc_list):
        """
        Finds the descriptor that describes the current image with the highest priority

        Args:
            desc_list: list of descriptor instances

        Returns:
            res_desc: Single descriptor instance
        """
        pos_desc = [desc for desc in desc_list if desc.describes(self.name)]
        if len(pos_desc) == 0:
            return None
        else:
            max_priority = 0
            res_desc = None
            for desc in pos_desc:
                if desc.priority > max_priority:
                    max_priority = desc.priority
                    res_desc = desc
            self.old_name = self.name
            self.name = str(res_desc)
            self.new_name = True
            self.bids_info = res_desc.get_bids_info()
            return res_desc


def get_all_med_name(path_dataset):
    """
    Creates a list 'medical_images' containing all possible medical images' names in the NIFD dataset

    Args:
        path_dataset: path to the NIFD dataset

    Returns:
        medical_images = ['t1_mprage', 't1_mprage_S3_DIS3D', ...]
    """
    import os

    subfolders = [f.path for f in os.scandir(path_dataset) if f.is_dir()]
    medical_images = []
    for sub in subfolders:
        subfolders2 = [f.path for f in os.scandir(sub) if f.is_dir()]
        for sub2 in subfolders2:
            medical_image_name = os.path.basename(sub2)
            if medical_image_name not in medical_images:
                medical_images.append(medical_image_name)
    return medical_images


def get_descriptors(path_root):
    """
    Loading the descriptors

    Args:
        path_root: path to the folder containing the 'config_dcm2bids.json' file

    Returns:
        descriptors: list of descriptor instances
    """
    import os
    import json

    with open(os.path.join(path_root, 'config_dcm2bids.json')) as f:
        data = json.load(f)

    descriptors = []
    for i in data['descriptions']:
        des = Descriptor(i)
        descriptors.append(des)

    return descriptors


def dict_conversion(medical_images, descriptors):
    """
    For each medical image in the dataset, find whether its name is described by one of the descriptor object.
    If it is, the descriptor with the highest priority is selected.
    The medical image is then added to the dictionnary as a key with its associated descriptor and its modality label.

    Args:
        medical_images: list containing every medical image name
        descriptors: list of descriptor instances

    Returns:
        equivalences: dictionary, equivalences['medical_image_name'] = (Descriptor_instance, modalityLabel)
    """
    equivalences = dict()

    for med_name in medical_images:
        old_name = Old_image_folder(med_name)
        desc = old_name.get_new_name(descriptors)
        if old_name.new_name:
            equivalences[old_name.old_name] = (desc, old_name.bids_info)

    return equivalences



# coding: utf8

import clinica.pipelines.engine as cpe


class T1VolumeCreateDartel(cpe.Pipeline):
    """T1VolumeCreateDartel - Create new Dartel template.

    Returns:
        A clinica pipeline object containing the T1VolumeCreateDartel pipeline.
    """
    def check_pipeline_parameters(self):
        """Check pipeline parameters."""
        from clinica.utils.group import check_group_label

        if 'group_id' not in self.parameters.keys():
            raise KeyError('Missing compulsory group_id key in pipeline parameter.')
        if 'dartel_tissues' not in self.parameters.keys():
            self.parameters['dartel_tissues'] = [1, 2, 3]

        check_group_label(self.parameters['group_id'])

    def check_custom_dependencies(self):
        """Check dependencies that can not be listed in the `info.json` file."""
        pass

    def get_input_fields(self):
        """Specify the list of possible inputs of this pipelines.

        Returns:
            A list of (string) input fields name.
        """

        return ['dartel_inputs']

    def get_output_fields(self):
        """Specify the list of possible outputs of this pipelines.

        Returns:
            A list of (string) output fields name.
        """

        return ['final_template_file', 'template_files', 'dartel_flow_fields']

    def build_input_node(self):
        """Build and connect an input node to the pipeline."""
        import os
        import sys
        from colorama import Fore
        import nipype.pipeline.engine as npe
        import nipype.interfaces.utility as nutil
        from clinica.utils.inputs import clinica_file_reader
        from clinica.utils.input_files import t1_volume_dartel_input_tissue
        from clinica.utils.exceptions import ClinicaException
        from clinica.utils.stream import cprint
        from clinica.utils.ux import print_groups_in_caps_directory, print_images_to_process, print_begin_image

        representative_output = os.path.join(self.caps_directory,
                                             'groups',
                                             'group-' + self.parameters['group_id'],
                                             't1',
                                             'group-' + self.parameters['group_id'] + '_template.nii.gz'
                                             )
        if os.path.exists(representative_output):
            cprint("%sDARTEL template for %s already exists. Currently, Clinica does not propose to overwrite outputs "
                   "for this pipeline.%s" % (Fore.YELLOW, self.parameters['group_id'], Fore.RESET))
            print_groups_in_caps_directory(self.caps_directory)
            sys.exit(0)

        # Check that there is at least 2 subjects
        if len(self.subjects) <= 1:
            raise ClinicaException('%sThis pipeline needs at least 2 images to create DARTEL template but '
                                   'Clinica only found %s.%s' % (Fore.RED, len(self.subjects), Fore.RESET))

        read_parameters_node = npe.Node(name="LoadingCLIArguments",
                                        interface=nutil.IdentityInterface(fields=self.get_input_fields(),
                                                                          mandatory_inputs=True))
        all_errors = []
        d_input = []
        for tissue_number in self.parameters['dartel_tissues']:
            try:
                current_file = clinica_file_reader(self.subjects,
                                                   self.sessions,
                                                   self.caps_directory,
                                                   t1_volume_dartel_input_tissue(tissue_number))
                d_input.append(current_file)
            except ClinicaException as e:
                all_errors.append(e)

        # Raise all errors if some happened
        if len(all_errors) > 0:
            error_message = 'Clinica faced errors while trying to read files in your BIDS or CAPS directories.\n'
            for msg in all_errors:
                error_message += str(msg)
            raise RuntimeError(error_message)

        # d_input is a list of size len(self.parameters['dartel_tissues'])
        #     Each element of this list is a list of size len(self.subjects)
        read_parameters_node.inputs.dartel_inputs = d_input

        if len(self.subjects):
            print_images_to_process(self.subjects, self.sessions)
            cprint('Computational time for DARTEL creation will depend on the number of images.')
            print_begin_image('group-' + self.parameters['group_id'])

        self.connect([
            (read_parameters_node, self.input_node, [('dartel_inputs', 'dartel_input_images')])
        ])

    def build_output_node(self):
        """Build and connect an output node to the pipeline."""
        import os.path as op
        import nipype.pipeline.engine as npe
        import nipype.interfaces.io as nio
        import re
        from clinica.utils.filemanip import zip_nii

        # Writing flowfields into CAPS
        # ============================
        write_flowfields_node = npe.MapNode(name='write_flowfields_node',
                                            iterfield=['container', 'flow_fields'],
                                            interface=nio.DataSink(infields=['flow_fields']))
        write_flowfields_node.inputs.base_directory = self.caps_directory
        write_flowfields_node.inputs.parameterization = False
        write_flowfields_node.inputs.container = ['subjects/' + self.subjects[i] + '/' + self.sessions[i] +
                                                  '/t1/spm/dartel/group-' + self.parameters['group_id']
                                                  for i in range(len(self.subjects))]
        write_flowfields_node.inputs.regexp_substitutions = [
            (r'(.*)_Template(\.nii(\.gz)?)$', r'\1\2'),
            (r'(.*)c1(sub-.*)(\.nii(\.gz)?)$', r'\1\2_segm-graymatter\3'),
            (r'(.*)c2(sub-.*)(\.nii(\.gz)?)$', r'\1\2_segm-whitematter\3'),
            (r'(.*)c3(sub-.*)(\.nii(\.gz)?)$', r'\1\2_segm-csf\3'),
            (r'(.*)c4(sub-.*)(\.nii(\.gz)?)$', r'\1\2_segm-bone\3'),
            (r'(.*)c5(sub-.*)(\.nii(\.gz)?)$', r'\1\2_segm-softtissue\3'),
            (r'(.*)c6(sub-.*)(\.nii(\.gz)?)$', r'\1\2_segm-background\3'),
            (r'(.*)r(sub-.*)(\.nii(\.gz)?)$', r'\1\2\3'),
            (r'(.*)_dartelinput(\.nii(\.gz)?)$', r'\1\2'),
            (r'(.*)flow_fields/u_(sub-.*)_segm-.*(\.nii(\.gz)?)$',
             r'\1\2_target-' + re.escape(self.parameters['group_id']) + r'_transformation-forward_deformation\3'),
            (r'trait_added', r'')
        ]

        # Writing templates into CAPS
        # ===========================
        write_template_node = npe.Node(nio.DataSink(), name='write_template_node')
        write_template_node.inputs.parameterization = False
        write_template_node.inputs.base_directory = self.caps_directory
        write_template_node.inputs.container = op.join('groups/group-' + self.parameters['group_id'], 't1')
        write_template_node.inputs.regexp_substitutions = [
            (r'(.*)final_template_file/.*(\.nii(\.gz)?)$',
             r'\1group-' + re.escape(self.parameters['group_id']) + r'_template\2'),
            (r'(.*)template_files/.*([0-9])(\.nii(\.gz)?)$',
             r'\1group-' + re.escape(self.parameters['group_id']) + r'_iteration-\2_template\3')
        ]

        self.connect([
            (self.output_node, write_flowfields_node, [(('dartel_flow_fields', zip_nii, True), 'flow_fields')]),
            (self.output_node, write_template_node, [(('final_template_file', zip_nii, True), 'final_template_file'),
                                                     (('template_files', zip_nii, True), 'template_files')])
        ])

    def build_core_nodes(self):
        """Build and connect the core nodes of the pipeline."""
        import nipype.interfaces.spm as spm
        import nipype.pipeline.engine as npe
        import nipype.interfaces.utility as nutil
        from clinica.utils.filemanip import unzip_nii
        from clinica.utils.spm import spm_standalone_is_available, use_spm_standalone

        if spm_standalone_is_available():
            use_spm_standalone()

        # Unzipping
        # =========
        unzip_node = npe.MapNode(nutil.Function(input_names=['in_file'],
                                                output_names=['out_file'],
                                                function=unzip_nii),
                                 name='unzip_node', iterfield=['in_file'])

        # DARTEL template
        # ===============
        dartel_template = npe.Node(spm.DARTEL(),
                                   name='dartel_template')

        # Connection
        # ==========
        self.connect([
            (self.input_node, unzip_node,    [('dartel_input_images', 'in_file')]),
            (unzip_node, dartel_template,    [('out_file', 'image_files')]),
            (dartel_template, self.output_node, [('dartel_flow_fields', 'dartel_flow_fields'),
                                                 ('final_template_file', 'final_template_file'),
                                                 ('template_files', 'template_files')])
        ])



# coding: utf-8

"""
Module for converting Tau PET of ADNI
"""


def convert_adni_tau_pet(source_dir, csv_dir, dest_dir, subjs_list=None):
    """
    Convert Tau PET images of ADNI into BIDS format

    Args:
        source_dir: path to the ADNI directory
        csv_dir: path to the clinical data directory
        dest_dir: path to the destination BIDS directory
        subjs_list: subjects list

    """

    import pandas as pd
    from os import path
    from clinica.utils.stream import cprint
    from clinica.iotools.converters.adni_to_bids.adni_utils import paths_to_bids
    from colorama import Fore

    if subjs_list is None:
        adni_merge_path = path.join(csv_dir, 'ADNIMERGE.csv')
        adni_merge = pd.read_csv(adni_merge_path, sep=',', low_memory=False)
        subjs_list = list(adni_merge.PTID.unique())

    cprint('Calculating paths of TAU PET images. Output will be stored in %s.' % path.join(dest_dir, 'conversion_info'))
    images = compute_tau_pet_paths(source_dir, csv_dir, dest_dir, subjs_list)
    cprint('Paths of TAU PET images found. Exporting images into BIDS ...')
    paths_to_bids(images, dest_dir, 'tau')
    cprint(Fore.GREEN + 'TAU PET conversion done.' + Fore.RESET)


def compute_tau_pet_paths(source_dir, csv_dir, dest_dir, subjs_list):
    """
    Compute the paths to Tau PET images

    Args:
        source_dir: path to the ADNI directory
        csv_dir: path to the clinical data directory
        dest_dir: path to the destination BIDS directory
        subjs_list: subjects list

    Returns: pandas Dataframe containing the path for each Tau PET image

    """

    import pandas as pd
    import os
    from os import path
    from clinica.iotools.converters.adni_to_bids.adni_utils import get_images_pet, find_image_path

    pet_tau_col = ['Phase', 'Subject_ID', 'VISCODE', 'Visit', 'Sequence', 'Scan_Date', 'Study_ID',
                   'Series_ID', 'Image_ID', 'Original']
    pet_tau_df = pd.DataFrame(columns=pet_tau_col)
    pet_tau_dfs_list = []

    # Loading needed .csv files
    tauqc = pd.read_csv(path.join(csv_dir, 'TAUQC.csv'), sep=',', low_memory=False)
    tauqc3 = pd.read_csv(path.join(csv_dir, 'TAUQC3.csv'), sep=',', low_memory=False)
    pet_meta_list = pd.read_csv(path.join(csv_dir, 'PET_META_LIST.csv'), sep=',', low_memory=False)

    for subj in subjs_list:

        # PET images metadata for subject
        subject_pet_meta = pet_meta_list[pet_meta_list['Subject'] == subj]

        if subject_pet_meta.empty:
            continue

        # QC for TAU PET images for ADNI 2
        tau_qc2_subj = tauqc[(tauqc.SCANQLTY == 1) & (tauqc.RID == int(subj[-4:]))]

        # QC for TAU PET images for ADNI 3
        tau_qc3_subj = tauqc3[(tauqc3.SCANQLTY == 1) & (tauqc3.RID == int(subj[-4:]))]

        # Concatenating visits in both QC files
        tau_qc_subj = pd.concat([tau_qc2_subj, tau_qc3_subj], axis=0, ignore_index=True, sort=False)
        tau_qc_subj.rename(columns={"SCANDATE": "EXAMDATE"}, inplace=True)

        sequences_preprocessing_step = ['AV1451 Co-registered, Averaged']
        subj_dfs_list = get_images_pet(subj, tau_qc_subj, subject_pet_meta, pet_tau_col, 'TAU-PET',
                                       sequences_preprocessing_step)
        if subj_dfs_list:
            pet_tau_dfs_list += subj_dfs_list

    if pet_tau_dfs_list:
        pet_tau_df = pd.concat(pet_tau_dfs_list, ignore_index=True)

    # Exceptions
    # ==========
    conversion_errors = [  # Multiple output images
                         ('098_S_4275', 'm84')]

    # Removing known exceptions from images to convert
    if not pet_tau_df.empty:
        error_ind = pet_tau_df.index[pet_tau_df.apply(lambda x: ((x.Subject_ID, x.VISCODE) in conversion_errors),
                                                      axis=1)]
        pet_tau_df.drop(error_ind, inplace=True)

    # Checking for images paths in filesystem
    images = find_image_path(pet_tau_df, source_dir, 'TAU', 'I', 'Image_ID')

    tau_csv_path = path.join(dest_dir, 'conversion_info')
    if not os.path.exists(tau_csv_path):
        os.mkdir(tau_csv_path)
    images.to_csv(path.join(tau_csv_path, 'tau_pet_paths.tsv'), sep='\t', index=False)

    return images



# coding: utf8


import clinica.pipelines.engine as cpe


class T1FreeSurferTemplate(cpe.Pipeline):
    """FreeSurfer Longitudinal template class

    Returns:
        A clinica pipeline object containing the T1FreeSurferTemplate pipeline.
    """

    @staticmethod
    def get_processed_images(caps_directory, subjects, sessions):
        import os
        from clinica.utils.inputs import clinica_file_reader
        from clinica.utils.input_files import T1_FS_T_DESTRIEUX
        from clinica.utils.longitudinal import get_long_id
        from clinica.utils.participant import get_unique_subjects
        import re

        [list_participant_id, list_list_session_ids] = get_unique_subjects(subjects, sessions)
        list_long_id = [get_long_id(list_session_ids) for list_session_ids in list_list_session_ids]

        image_ids = []
        if os.path.isdir(caps_directory):
            t1_freesurfer_files = clinica_file_reader(
                list_participant_id, list_long_id,
                caps_directory, T1_FS_T_DESTRIEUX, False
            )
            image_ids = [re.search(r'(sub-[a-zA-Z0-9]+)_(long-[a-zA-Z0-9]+)', file).group()
                         for file in t1_freesurfer_files]
        return image_ids

    def check_pipeline_parameters(self):
        """Check pipeline parameters."""
        pass

    def check_custom_dependencies(self):
        """Check dependencies that cannot be listed in `info.json`
        """
        pass

    def get_input_fields(self):
        """Specify the list of possible inputs of this pipeline.

        Note:
            The list of inputs of the T1FreeSurferTemplate pipeline is:
                * participant_id (str): Participant ID (e.g. "sub-CLNC01")
                * list_session_ids (List[str]: List of session IDs associated to `participant_id`
                    (e.g. ["ses-M00"] or ["ses-M00", "ses-M18", "ses-M36"])

        Returns:
            A list of (string) input fields name.
        """
        return ['participant_id', 'list_session_ids']

    def get_output_fields(self):
        """Specify the list of possible outputs of this pipeline.

        Note:
            The list of inputs of the T1FreeSurferTemplate pipeline is:
                * image_id (str): Image ID (e.g. "sub-CLNC01_long-M00M18")

        Returns:
            A list of (string) output fields name.
        """
        return ['image_id']

    def build_input_node(self):
        """Build and connect an input node to the pipeline."""
        import os
        from colorama import Fore

        import nipype.interfaces.utility as nutil
        import nipype.pipeline.engine as npe

        from clinica.utils.exceptions import ClinicaException, ClinicaCAPSError
        from clinica.utils.filemanip import extract_subjects_sessions_from_filename
        from clinica.utils.inputs import clinica_file_reader
        from clinica.utils.input_files import T1_FS_DESTRIEUX
        from clinica.utils.longitudinal import get_long_id, read_sessions, get_participants_long_id
        from clinica.utils.participant import get_unique_subjects, unique_subjects_sessions_to_subjects_sessions
        from clinica.utils.stream import cprint
        from .longitudinal_utils import extract_participant_long_ids_from_filename, save_part_sess_long_ids_to_tsv

        # Display image(s) already present in CAPS folder
        # ===============================================
        output_ids = self.get_processed_images(self.caps_directory, self.subjects, self.sessions)
        processed_participants, processed_long_sessions = extract_participant_long_ids_from_filename(output_ids)
        if len(processed_participants) > 0:
            cprint("%sClinica found %s participant(s) already processed in CAPS directory:%s" %
                   (Fore.YELLOW, len(processed_participants), Fore.RESET))
            for p_id, l_id in zip(processed_participants, processed_long_sessions):
                cprint("%s\t%s | %s%s" % (Fore.YELLOW, p_id, l_id, Fore.RESET))
            if self.overwrite_caps:
                output_folder = "<CAPS>/subjects/<participant_id>/<long_id>/freesurfer_unbiased_template/"
                cprint("%s\nOutput folders in %s will be recreated.\n%s" % (Fore.YELLOW, output_folder, Fore.RESET))
            else:
                cprint("%s\nParticipant(s) will be ignored by Clinica.\n%s" % (Fore.YELLOW, Fore.RESET))
                input_ids = [p_id + '_' + s_id for p_id, s_id in
                             zip(self.subjects, self.sessions)]
                processed_sessions_per_participant = [
                    read_sessions(self.caps_directory, p_id, l_id)
                    for (p_id, l_id) in zip(processed_participants, processed_long_sessions)
                ]
                participants, sessions = unique_subjects_sessions_to_subjects_sessions(
                    processed_participants,
                    processed_sessions_per_participant)
                processed_ids = [p_id + '_' + s_id for p_id, s_id in zip(participants, sessions)]
                to_process_ids = list(set(input_ids) - set(processed_ids))
                self.subjects, self.sessions = extract_subjects_sessions_from_filename(to_process_ids)

        # Check that t1-freesurfer has run on the CAPS directory
        try:
            clinica_file_reader(self.subjects,
                                self.sessions,
                                self.caps_directory,
                                T1_FS_DESTRIEUX)
        except ClinicaException as e:
            err_msg = 'Clinica faced error(s) while trying to read files in your CAPS directory.\n' + str(e)
            raise ClinicaCAPSError(err_msg)

        # Save subjects to process in <WD>/<Pipeline.name>/participants.tsv
        folder_participants_tsv = os.path.join(self.base_dir, self.name)
        long_ids = get_participants_long_id(self.subjects, self.sessions)
        save_part_sess_long_ids_to_tsv(self.subjects, self.sessions, long_ids, folder_participants_tsv)

        [list_participant_id, list_list_session_ids] = get_unique_subjects(self.subjects, self.sessions)
        list_long_id = [get_long_id(list_session_ids) for list_session_ids in list_list_session_ids]

        def print_images_to_process(unique_part_list, per_part_session_list, list_part_long_id):
            cprint('The pipeline will be run on the following %s participant(s):' % len(unique_part_list))
            for (part_id, list_sess_id, list_id) in zip(unique_part_list, per_part_session_list, list_part_long_id):
                sessions_participant = ', '.join(s_id for s_id in list_sess_id)
                cprint("\t%s | %s | %s" % (part_id, sessions_participant, list_id))

        if len(self.subjects):
            # TODO: Generalize long IDs to the message display
            print_images_to_process(list_participant_id, list_list_session_ids, list_long_id)
            cprint('List available in %s' % os.path.join(folder_participants_tsv, 'participants.tsv'))
            cprint('The pipeline will last approximately 10 hours per participant.')

        read_node = npe.Node(name="ReadingFiles",
                             iterables=[
                                 ('participant_id', list_participant_id),
                                 ('list_session_ids', list_list_session_ids),
                             ],
                             synchronize=True,
                             interface=nutil.IdentityInterface(
                                 fields=self.get_input_fields())
                             )
        self.connect([
            (read_node, self.input_node, [('participant_id', 'participant_id')]),
            (read_node, self.input_node, [('list_session_ids', 'list_session_ids')]),
        ])

    def build_output_node(self):
        """Build and connect an output node to the pipeline."""
        import nipype.pipeline.engine as npe
        import nipype.interfaces.utility as nutil
        from .t1_freesurfer_template_utils import save_to_caps
        import os

        save_to_caps = npe.Node(interface=nutil.Function(
            input_names=['source_dir', 'image_id', 'list_session_ids', 'caps_dir', 'overwrite_caps'],
            output_names=['image_id'],
            function=save_to_caps),
            name='SaveToCaps')
        save_to_caps.inputs.source_dir = os.path.join(self.base_dir, self.name, 'ReconAll')
        save_to_caps.inputs.caps_dir = self.caps_directory
        save_to_caps.inputs.overwrite_caps = False

        self.connect([
            (self.input_node, save_to_caps, [('list_session_ids', 'list_session_ids')]),
            (self.output_node, save_to_caps, [('image_id', 'image_id')]),
        ])

    def build_core_nodes(self):
        """Build and connect the core nodes of the pipeline."""
        import os
        import nipype.interfaces.utility as nutil
        import nipype.pipeline.engine as npe
        from .t1_freesurfer_template_utils import init_input_node, run_recon_all_base, move_subjects_dir_to_source_dir

        # Nodes declaration
        # =================
        # Initialize the pipeline
        init_input = npe.Node(
            interface=nutil.Function(
                input_names=['caps_dir', 'participant_id', 'list_session_ids', 'output_dir'],
                output_names=['image_id', 'subjects_dir', 'flags'],
                function=init_input_node),
            name='0-InitPipeline')
        init_input.inputs.caps_dir = self.caps_directory
        init_input.inputs.output_dir = os.path.join(self.base_dir, self.name, 'ReconAll')

        # Run recon-all command
        recon_all = npe.Node(
            interface=nutil.Function(
                input_names=['subjects_dir', 'subject_id', 'flags', 'directive'],
                output_names=['subject_id'],
                function=run_recon_all_base),
            name='1-SegmentationReconAll')
        recon_all.inputs.directive = '-all'

        # Move $SUBJECT_DIR to source_dir (1 time point case)
        move_subjects_dir = npe.Node(
            interface=nutil.Function(
                input_names=['subjects_dir', 'source_dir', 'subject_id'],
                output_names=['subject_id'],
                function=move_subjects_dir_to_source_dir),
            name='2-MoveSubjectsDir')
        move_subjects_dir.inputs.source_dir = os.path.join(self.base_dir, self.name, 'ReconAll')

        # Connections
        # ===========
        self.connect([
            # Initialize the pipeline
            (self.input_node, init_input, [('participant_id', 'participant_id')]),
            (self.input_node, init_input, [('list_session_ids', 'list_session_ids')]),
            # Run recon-all command
            (init_input, recon_all, [('subjects_dir', 'subjects_dir'),
                                     ('image_id', 'subject_id'),
                                     ('flags',  'flags')]),
            # Move $SUBJECT_DIR to source_dir (1 time point case)
            (init_input, move_subjects_dir, [('subjects_dir', 'subjects_dir')]),
            (recon_all, move_subjects_dir, [('subject_id', 'subject_id')]),
            # Output node
            (move_subjects_dir, self.output_node, [('subject_id', 'image_id')]),
        ])



# coding: utf8
"""
This module contains FreeSurfer utilities.
"""


def extract_image_id_from_longitudinal_segmentation(freesurfer_id):
    """Extract image ID from longitudinal segmentation folder.

    This function will extract participant, session and longitudinal ID from `freesurfer_id`.

    Example:
        >>> from clinica.utils.freesurfer import extract_image_id_from_longitudinal_segmentation
        >>> extract_image_id_from_longitudinal_segmentation('sub-CLNC01_ses-M00')
        image_id(participant_id='sub-CLNC01', session_id='ses-M00', long_id='')
        >>> extract_image_id_from_longitudinal_segmentation('sub-CLNC01_long-M0018')
        image_id(participant_id='sub-CLNC01', session_id='', long_id='long-M0018')
        >>> extract_image_id_from_longitudinal_segmentation('sub-CLNC01_ses-M00.long.sub-CLNC01_long-M00M18')
        image_id(participant_id='sub-CLNC01', session_id='ses-M00', long_id='long-M00M18')
    """
    from collections import namedtuple
    image_id = namedtuple('image_id', ['participant_id', 'session_id', 'long_id'])

    # Case 'sub-CLNC01_ses-M00.long.sub-CLNC01_long-M00M18'
    if '.long.' in freesurfer_id:
        participant_id = freesurfer_id.split('.long.')[0].split('_')[0]
        session_id = freesurfer_id.split('.long.')[0].split('_')[1]
        long_id = freesurfer_id.split('.long.')[1].split('_')[1]
    # Case 'sub-CLNC01_long-M00M18'
    elif 'long-' in freesurfer_id:
        participant_id = freesurfer_id.split('_')[0]
        session_id = ''
        long_id = freesurfer_id.split('_')[1]
    # Case 'sub-CLNC01_ses-M00'
    else:
        participant_id = freesurfer_id.split('_')[0]
        session_id = freesurfer_id.split('_')[1]
        long_id = ''

    return image_id(participant_id, session_id, long_id)


def get_secondary_stats(stats_filename, info_type):
    """Read the 'secondary' statistical info from .stats file

    Extract the information from .stats file that is commented out
    (lines starting with '# Measure' prefix) and does not appear in the
    table at the end of the document.

    Args:
        stats_filename (string): path to the .stats file
        info_type (string): 'volume', 'thickness' or 'area'

    Returns:
        secondary_stats_dict (dictionary of float): keys are regions of
            the brain, associated values are the corresponding
            volume/thickness/area depending on the input info type
    """
    # initialise structure containing the secondary statistical info
    secondary_stats_dict = dict()

    # currently no additional information is provided by .stats file for
    # the mean curvature
    if info_type != 'meancurv':
        # define how lines are supposed to end in the stats file, depending
        # on the type of information that is searched for
        endline_dict = dict()
        endline_dict['volume'] = 'mm^3'
        endline_dict['thickness'] = 'mm'
        endline_dict['area'] = 'mm^2'

        # define keywords that are supposed to appear in commented lines
        # containing statistical information
        info_keyword_dict = dict()
        info_keyword_dict['volume'] = ['volume', 'Volume']
        info_keyword_dict['area'] = ['area', 'Area']
        info_keyword_dict['thickness'] = ['thickness', 'Thickness']

        # read stats file line by line and only keep relevant lines
        with open(stats_filename, 'r') as stats_file:
            stats = stats_file.read()
        stats_line_list = stats.splitlines()
        for stats_line in stats_line_list:
            startswith_condition = stats_line.startswith('# Measure')
            endswith_condition = stats_line.endswith(endline_dict[info_type])
            if (startswith_condition) and (endswith_condition):
                stats_line_word_list = stats_line.replace(',', '').split()
                # sanity check: make sure any sensible variation of
                # 'volume', 'thickness' or 'area' appears inside the line
                if any(x in stats_line_word_list for x in info_keyword_dict[info_type]):
                    # add info
                    info_region = stats_line_word_list[2]
                    info_value = stats_line_word_list[-2]
                    secondary_stats_dict[info_region] = info_value

    return secondary_stats_dict


def generate_regional_measures(
        segmentation_path, subject_id, output_dir=None):
    """
    Read stats files located in
    <segmentation_path>/<subject_id>/stats/*.stats
    and generate TSV files in <segmentation_path>/regional_measures
    folder.

    Note: the .stats files contain both 1) a table with statistical
    information (e.g., structure volume) and 2) 'secondary' statistical
    information with all lines starting with the sentence '# Measure'.
    The .tsv files return the relevant statistical information from both
    sources.

    Args:
        segmentation_path (string): Path to the FreeSurfer segmentation.
        subject_id (string): Subject ID in the form sub-CLNC01_ses-M00, sub-CLNC01_long-M00M18 or
            sub-CLNC01_ses-M00.long.sub-CLNC01_long-M00M18
        output_dir (string): folder where the .tsv stats files will be
            stored. Will be [path_segmentation]/regional_measures if no
            dir is provided by the user
    """
    import os
    import errno
    import pandas
    from clinica.utils.freesurfer import write_tsv_file

    image_id = extract_image_id_from_longitudinal_segmentation(subject_id)
    prefix = image_id.participant_id
    if image_id.session_id:
        prefix = prefix + '_' + image_id.session_id
    if image_id.long_id:
        prefix = prefix + '_' + image_id.long_id

    stats_folder = os.path.join(os.path.expanduser(segmentation_path), subject_id, 'stats')

    if not os.path.isdir(stats_folder):
        raise IOError("Image %s does not contain FreeSurfer segmentation" % prefix.replace('_', ' | '))

    if not output_dir:
        output_dir = os.path.join(segmentation_path, 'regional_measures')
    try:
        os.makedirs(output_dir)
    except OSError as exception:
        # if dest_dir exists, go on, if its other error, raise
        if exception.errno != errno.EEXIST:
            raise

    # Generate TSV files for parcellation files
    #
    # Columns in ?h.BA.stats, ?h.aparc.stats or ?h.aparc.a2009s.stats file
    columns_parcellation = [
        'StructName', 'NumVert', 'SurfArea', 'GrayVol', 'ThickAvg', 'ThickStd',
        'MeanCurv', 'GausCurv', 'FoldInd', 'CurvInd'
    ]
    hemi_dict = {
        'left': 'lh',
        'right': 'rh'
    }
    atlas_dict = {
        'desikan': 'aparc',
        'destrieux': 'aparc.a2009s',
        'ba': 'BA_exvivo'
    }
    info_dict = {
        'volume': 'GrayVol',
        'thickness': 'ThickAvg',
        'area': 'SurfArea',
        'meancurv': 'MeanCurv'
    }
    for atlas in ('desikan', 'destrieux', 'ba'):
        stats_filename_dict = dict()
        df_dict = dict()
        # read both left and right .stats files
        for hemi in ('left', 'right'):
            stats_filename_dict[hemi] = os.path.join(
                stats_folder,
                '{0}.{1}.stats'.format(hemi_dict[hemi], atlas_dict[atlas]))
            df_dict[hemi] = pandas.read_csv(
                stats_filename_dict[hemi],
                names=columns_parcellation,
                comment='#', header=None, delimiter='\s+', dtype=str)
        # generate .tsv from 1) the table in .stats file and 2) the
        # secondary (commented out) information common to both 'left'
        # and 'right' .stats file
        for info in ('volume', 'thickness', 'area', 'meancurv'):
            # Secondary information (common to 'left' and 'right')
            secondary_stats_dict = get_secondary_stats(stats_filename_dict['left'], info)
            # Join primary and secondary information
            key_list = (list('lh_'+df_dict['left']['StructName']) +
                        list('rh_'+df_dict['right']['StructName']) +
                        list(secondary_stats_dict.keys()))
            col_name = info_dict[info]
            value_list = (list(df_dict['left'][col_name]) +
                          list(df_dict['right'][col_name]) +
                          list(secondary_stats_dict.values()))
            # Write .tsv
            write_tsv_file(os.path.join(output_dir, '{0}_parcellation-{1}_{2}.tsv'.format(prefix, atlas, info)),
                           key_list, info, value_list)

    # Generate TSV files for segmentation files
    #
    # Columns in aseg.stats or wmparc.stats file
    columns_segmentation = [
        'Index', 'SegId', 'NVoxels', 'Volume_mm3', 'StructName', 'normMean',
        'normStdDev', 'normMin', 'normMax', 'normRange']

    # Parsing aseg.stats
    stats_filename = os.path.join(stats_folder, 'aseg.stats')
    df = pandas.read_csv(
        stats_filename,
        comment='#', header=None, delimiter='\s+', dtype=str,
        names=columns_segmentation)
    secondary_stats_dict = get_secondary_stats(stats_filename, 'volume')
    key_list = list(df['StructName'])+list(secondary_stats_dict.keys())
    value_list = list(df['Volume_mm3'])+list(secondary_stats_dict.values())
    write_tsv_file(os.path.join(output_dir, prefix + '_segmentationVolumes.tsv'),
                   key_list, 'volume', value_list)

    # Parsing wmparc.stats
    stats_filename = os.path.join(stats_folder, 'wmparc.stats')
    df = pandas.read_csv(
        stats_filename,
        comment='#', header=None, delimiter='\s+', dtype=str,
        names=columns_segmentation)
    secondary_stats_dict = get_secondary_stats(stats_filename, 'volume')
    key_list = list(df['StructName'])+list(secondary_stats_dict.keys())
    value_list = list(df['Volume_mm3'])+list(secondary_stats_dict.values())
    write_tsv_file(os.path.join(output_dir, prefix + '_parcellation-wm_volume.tsv'),
                   key_list, 'volume', value_list)


def write_tsv_file(out_filename, name_list, scalar_name, scalar_list):
    """Write a .tsv file with list of keys and values

    Args:
        out_filename (string): name of the .tsv file
        name_list (list of string): list of keys
        scalar_name (string): 'volume', 'thickness', 'area' or
            'meanCurv'. Not used for now. Might be used as part of a
            pandas data frame
        scalar_list (list of float): list of values corresponding to the keys
    """
    import pandas
    import warnings

    try:
        data = pandas.DataFrame({
            'label_name': name_list,
            'label_value': scalar_list})
        data.to_csv(out_filename, sep='\t', index=False, encoding='utf-8')
    except Exception as exception:
        warnings.warn("Impossible to save {0} file".format(out_filename))
        raise exception

    return out_filename


def check_flags(in_t1w, recon_all_args):
    """Check `recon_all_args` flags for `in_t1w` image.

    Currently, this function only adds '-cw256' if the FOV of `in_t1w` is greater than 256.
    """
    import nibabel as nib
    # from clinica.utils.stream import cprint

    f = nib.load(in_t1w)
    voxel_size = f.header.get_zooms()
    t1_size = f.header.get_data_shape()
    if (voxel_size[0] * t1_size[0] > 256) or \
            (voxel_size[1] * t1_size[1] > 256) or \
            (voxel_size[2] * t1_size[2] > 256):
        # cprint("Setting MRI Convert to crop images to 256 FOV for %s file." % in_t1w)
        optional_flag = ' -cw256'
    else:
        # cprint("No need to add -cw256 flag for %s file." % in_t1w)
        optional_flag = ''
    flags = "{0}".format(recon_all_args) + optional_flag

    return flags



# coding: utf8


import abc
import os.path as path

import numpy as np
from pandas.io import parsers

from clinica.utils.stream import cprint
from clinica.pipelines.machine_learning import base
import clinica.pipelines.machine_learning.voxel_based_io as vbio
import clinica.pipelines.machine_learning.vertex_based_io as vtxbio
import clinica.pipelines.machine_learning.region_based_io as rbio
import clinica.pipelines.machine_learning.tsv_based_io as tbio
import clinica.pipelines.machine_learning.ml_utils as utils




__author__ = "Jorge Samper-Gonzalez"
__copyright__ = "Copyright 2016-2019 The Aramis Lab Team"
__credits__ = ["Jorge Samper-Gonzalez", "Simona Bottani"]
__license__ = "See LICENSE.txt file"
__version__ = "0.1.0"
__maintainer__ = "Jorge Samper-Gonzalez"
__email__ = "jorge.samper-gonzalez@inria.fr"
__status__ = "Development"


class CAPSInput(base.MLInput):

    def __init__(self, input_params):

        super().__init__(input_params)

        self._images = None

        subjects_visits = parsers.read_csv(self._input_params['subjects_visits_tsv'], sep='\t')
        if list(subjects_visits.columns.values) != ['participant_id', 'session_id']:
            raise Exception('Subjects and visits file is not in the correct format.')
        self._subjects = list(subjects_visits.participant_id)
        self._sessions = list(subjects_visits.session_id)

        diagnoses = parsers.read_csv(self._input_params['diagnoses_tsv'], sep='\t')
        if 'diagnosis' not in list(diagnoses.columns.values):
            raise Exception('Diagnoses file is not in the correct format.')
        self._diagnoses = list(diagnoses.diagnosis)

        if self._input_params['image_type'] not in ['T1', 'fdg', 'av45', 'pib', 'flute', 'dwi']:
            raise Exception("Incorrect image type. It must be one of the values 'T1', 'fdg', 'av45', "
                            "'pib', 'flute' or 'dwi'")

        if self._input_params['precomputed_kernel'] is not None:
            if type(self._input_params['precomputed_kernel']) == np.ndarray:
                if self._input_params['precomputed_kernel'].shape == (len(self._subjects), len(self._subjects)):
                    self._kernel = self._input_params['precomputed_kernel']
                else:
                    raise Exception("""Precomputed kernel provided is not in the correct format.
                    It must be a numpy.ndarray object with number of rows and columns equal to the number of subjects,
                    or a filename to a numpy txt file containing an object with the described format.""")
            elif type(self._input_params['precomputed_kernel'] == str):
                self._kernel = np.loadtxt(self._input_params['precomputed_kernel'])
            else:
                raise Exception("""Precomputed kernel provided is not in the correct format.
                It must be a numpy.ndarray object with number of rows and columns equal to the number of subjects,
                or a filename to a numpy txt file containing an object with the described format.""")

    @abc.abstractmethod
    def get_images(self):
        """

        Returns: a list of filenames

        """
        pass

    @abc.abstractmethod
    def get_x(self):
        """

        Returns: a numpy 2d-array.

        """
        pass

    def get_y(self):
        """

        Returns: a list of integers. Each integer represents a class.

        """
        if self._y is not None:
            return self._y

        unique = sorted(list(set(self._diagnoses)))
        self._y = np.array([unique.index(x) for x in self._diagnoses])
        return self._y

    def get_kernel(self, kernel_function=utils.gram_matrix_linear, recompute_if_exists=False):
        """

        Returns: a numpy 2d-array.

        """
        if self._kernel is not None and not recompute_if_exists:
            return self._kernel

        if self._x is None:
            self.get_x()

        cprint("Computing kernel ...")
        self._kernel = kernel_function(self._x)
        cprint("Kernel computed")
        return self._kernel

    def save_kernel(self, output_dir):
        """

        Args:
            output_dir:

        Returns:

        """
        if self._kernel is not None:
            filename = path.join(output_dir, 'kernel.txt')
            np.savetxt(filename, self._kernel)
            return filename
        raise Exception("Unable to save the kernel. Kernel must have been computed before.")

    @abc.abstractmethod
    def save_weights_as_nifti(self, weights, output_dir):
        pass

    @staticmethod
    def get_default_parameters():

        parameters_dict = {'caps_directory': None,
                           'subjects_visits_tsv': None,
                           'diagnoses_tsv': None,
                           'group_id': None,
                           'image_type': None,
                           'precomputed_kernel': None}

        return parameters_dict


class CAPSVoxelBasedInput(CAPSInput):

    def __init__(self, input_params):

        super().__init__(input_params)

        self._orig_shape = None
        self._data_mask = None

        if self._input_params['modulated'] not in ['on', 'off']:
            raise Exception("Incorrect modulation parameter. It must be one of the values 'on' or 'off'")

    def get_images(self):
        """

        Returns: a list of filenames

        """
        if self._images is not None:
            return self._images

        if self._input_params['image_type'] == 'T1':
            fwhm = '' if self._input_params['fwhm'] == 0 else '_fwhm-%dmm' % int(self._input_params['fwhm'])

            self._images = [path.join(self._input_params['caps_directory'], 'subjects', self._subjects[i],
                                      self._sessions[i], 't1/spm/dartel/group-' + self._input_params['group_id'],
                                      '%s_%s_T1w_segm-graymatter_space-Ixi549Space_modulated-%s%s_probability.nii.gz'
                                      % (self._subjects[i], self._sessions[i], self._input_params['modulated'], fwhm))
                            for i in range(len(self._subjects))]
        else:
            pvc = '' if self._input_params['pvc'] is None else '_pvc-%s' % self._input_params['pvc']
            fwhm = '' if self._input_params['fwhm'] == 0 else '_fwhm-%dmm' % int(self._input_params['fwhm'])
            suvr = 'pons' if self._input_params['image_type'] == 'fdg' else 'cerebellumPons'

            self._images = [path.join(self._input_params['caps_directory'], 'subjects', self._subjects[i],
                                      self._sessions[i], 'pet/preprocessing/group-' + self._input_params['group_id'],
                                      '%s_%s_task-rest_acq-%s_pet_space-Ixi549Space%s_suvr-%s_mask-brain%s_pet.nii.gz'
                                      % (self._subjects[i], self._sessions[i], self._input_params['image_type'], pvc,
                                         suvr, fwhm))
                            for i in range(len(self._subjects))]

        for image in self._images:
            if not path.exists(image):
                raise Exception("File %s doesn't exists." % image)

        return self._images

    def get_x(self):
        """

        Returns: a numpy 2d-array.

        """
        if self._x is not None:
            return self._x

        cprint('Loading ' + str(len(self.get_images())) + ' subjects')
        self._x, self._orig_shape, self._data_mask = vbio.load_data(self._images, mask=self._input_params['mask_zeros'])
        cprint('Subjects loaded')

        return self._x

    def save_weights_as_nifti(self, weights, output_dir):

        if self._images is None:
            self.get_images()

        output_filename = path.join(output_dir, 'weights.nii.gz')
        data = vbio.revert_mask(weights, self._data_mask, self._orig_shape)
        vbio.weights_to_nifti(data, self._images[0], output_filename)

    @staticmethod
    def get_default_parameters():

        parameters_dict = super(CAPSVoxelBasedInput, CAPSVoxelBasedInput).get_default_parameters()

        new_parameters = {'fwhm': 0,
                          'modulated': "on",
                          'pvc': None,
                          'mask_zeros': True}

        parameters_dict.update(new_parameters)

        return parameters_dict


class CAPSRegionBasedInput(CAPSInput):

    def __init__(self, input_params):

        super().__init__(input_params)

        if self._input_params['atlas'] not in ['AAL2', 'Neuromorphometrics', 'AICHA', 'LPBA40', 'Hammers']:
            raise Exception("Incorrect atlas name. It must be one of the values 'AAL2', 'Neuromorphometrics', "
                            "'AICHA', 'LPBA40', 'Hammers' ")

    def get_images(self):
        """

        Returns: a list of filenames

        """
        if self._images is not None:
            return self._images

        if self._input_params['image_type'] == 'T1':
            self._images = [path.join(self._input_params['caps_directory'], 'subjects', self._subjects[i],
                                      self._sessions[i], 't1/spm/dartel/group-' + self._input_params['group_id'],
                                      'atlas_statistics/', '%s_%s_T1w_space-%s_map-graymatter_statistics.tsv'
                                      % (self._subjects[i], self._sessions[i], self._input_params['atlas']))
                            for i in range(len(self._subjects))]
        else:
            pvc = '' if self._input_params['pvc'] is None else '_pvc-%s' % self._input_params['pvc']
            suvr = 'pons' if self._input_params['image_type'] == 'fdg' else 'cerebellumPons'

            self._images = [path.join(self._input_params['caps_directory'], 'subjects', self._subjects[i],
                                      self._sessions[i], 'pet/preprocessing/group-' + self._input_params['group_id'],
                                      'atlas_statistics', '%s_%s_task-rest_acq-%s_pet_space-%s%s_suvr-%s_statistics.tsv'
                                      % (self._subjects[i], self._sessions[i], self._input_params['image_type'],
                                         self._input_params['atlas'], pvc, suvr))
                            for i in range(len(self._subjects))]

        for image in self._images:
            if not path.exists(image):
                raise Exception("File %s doesn't exists." % image)

        return self._images

    def get_x(self):
        """

        Returns: a numpy 2d-array.

        """
        if self._x is not None:
            return self._x

        cprint('Loading ' + str(len(self.get_images())) + ' subjects')
        self._x = rbio.load_data(self._images, self._subjects)
        cprint('Subjects loaded')

        return self._x

    def save_weights_as_nifti(self, weights, output_dir):
        """

        Args:
            weights:
            output_dir:

        Returns:

        """

        output_filename = path.join(output_dir, 'weights.nii.gz')
        rbio.weights_to_nifti(weights, self._input_params['atlas'], output_filename)

    @staticmethod
    def get_default_parameters():

        parameters_dict = super(CAPSRegionBasedInput, CAPSRegionBasedInput).get_default_parameters()

        new_parameters = {'atlas': None,
                          'pvc': None,
                          'mask_zeros': True}

        parameters_dict.update(new_parameters)

        return parameters_dict


class CAPSVertexBasedInput(CAPSInput):

    def __init__(self, input_params):

        super().__init__(input_params)

    def get_images(self):
        import os
        """
        returns list of filnames
        """

        if self._images is not None:
            return self._images

        if self._input_params['image_type'] == 'fdg' and self._images is None:
            self._images = []
            hemi = ['lh', 'rh']
            for i in range(len(self._subjects)):
                self._images.append([os.path.join(self._input_params['caps_directory'], 'subjects', self._subjects[i],
                                                  self._sessions[i], 'pet', 'surface', self._subjects[i] + '_' +
                                                  self._sessions[i] + '_task-rest_acq-fdg_pet_space-fsaverage_'
                                                                      'suvr-pons_pvc-iy_hemi-' + h + '_fwhm-' +
                                                  str(self._input_params['fwhm']) + '_projection.mgh') for h in hemi])
            missing_files = []
            missing_files_string_error = ''
            for img in self._images:
                for side in img:
                    if not os.path.exists(side):
                        missing_files.append(side)
                        missing_files_string_error += side + '\n'
            if len(missing_files) > 0:
                raise IOError('Could not find the following files : \n' + missing_files_string_error
                              + '\n' + str(len(missing_files)) + ' files missing')
        return self._images

    def get_x(self):
        """
        Returns numpy 2D array
        """

        if self._x is not None:
            return self._x

        cprint('Loading ' + str(len(self.get_images())) + ' subjects')
        self._x = vtxbio.load_data(self._images)
        cprint(str(len(self._x)) + ' subjects loaded')
        return self._x

    def save_weights_as_datasurface(self, weights, output_dir):
        import numpy as np
        import nibabel as nib
        import os

        if self._images is None:
            self.get_images()

        sample = nib.load(self._images[0][0])

        infinite_norm = np.max(np.abs(weights))

        left_hemi_data = np.atleast_3d(np.divide(weights[:np.int(weights.size / 2)], infinite_norm))
        left_hemi_mgh = nib.MGHImage(left_hemi_data, affine=sample.affine, header=sample.header)
        nib.save(left_hemi_mgh, os.path.join(output_dir, 'weights_lh.mgh'))

        right_hemi_data = np.atleast_3d(np.divide(weights[np.int(weights.size/2):], infinite_norm))
        right_hemi_mgh = nib.MGHImage(right_hemi_data, affine=sample.affine, header=sample.header)
        nib.save(right_hemi_mgh, os.path.join(output_dir, 'weights_rh.mgh'))
        pass

    def save_weights_as_nifti(self, weights, output_dir):
        pass

    @staticmethod
    def get_default_parameters():

        parameters_dict = super(CAPSVertexBasedInput, CAPSVertexBasedInput).get_default_parameters()

        new_parameters = {'fwhm': 0}

        parameters_dict.update(new_parameters)

        return parameters_dict


class CAPSTSVBasedInput(CAPSInput):

    def __init__(self, input_params):

        super().__init__(input_params)

        if self._input_params['atlas'] not in ['AAL2', 'Neuromorphometrics', 'AICHA', 'LPBA40', 'Hammers']:
            raise Exception("Incorrect atlas name. It must be one of the values 'AAL2', 'Neuromorphometrics', "
                            "'AICHA', 'LPBA40', 'Hammers' ")

    def get_images(self):
        """

        Returns: string

        """

        pass

    def get_x(self):
        """

        Returns: a numpy 2d-array.

        """

        # if self._x is not None:
        #    return self._x

        cprint('Loading TSV subjects')
        string = str('group-' + self._input_params['group_id'] + '_T1w_space-' + self._input_params['atlas'] +
                     '_map-graymatter')

        self._x = tbio.load_data(string, self._input_params['caps_directory'], self._subjects, self._sessions,
                                 self._input_params['dataset'])

        cprint('Subjects loaded')

        return self._x

    def save_weights_as_nifti(self, weights, output_dir):
        """

        Args:
            weights:
            output_dir:

        Returns:

        """

        # output_filename = path.join(output_dir, 'weights.nii.gz')

        # rbio.weights_to_nifti(weights, self._input_params['atlas'], output_filename)
        pass

    @staticmethod
    def get_default_parameters():

        parameters_dict = super(CAPSTSVBasedInput, CAPSTSVBasedInput).get_default_parameters()

        new_parameters = {'atlas': None,
                          'pvc': None,
                          'dataset': None}

        parameters_dict.update(new_parameters)

        return parameters_dict


class CAPSVoxelBasedInputREGSVM(CAPSVoxelBasedInput):

    def get_images(self):
        """

        Returns: a list of filenames

        """
        if self._images is not None:
            return self._images

        if self._input_params['image_type'] == 'T1':
            fwhm = '' if self._input_params['fwhm'] == 0 else '_fwhm-%dmm' % int(self._input_params['fwhm'])

            self._images = [path.join(self._input_params['caps_directory'],
                                      'regul_%s_%s_T1w_segm-graymatter_space-Ixi549Space_modulated-%s%s_probability.nii'
                                      % (self._subjects[i], self._sessions[i], self._input_params['modulated'], fwhm))
                            for i in range(len(self._subjects))]
        else:
            pvc = '' if self._input_params['pvc'] is None else '_pvc-%s' % self._input_params['pvc']
            fwhm = '' if self._input_params['fwhm'] == 0 else '_fwhm-%dmm' % int(self._input_params['fwhm'])
            suvr = 'pons' if self._input_params['image_type'] == 'fdg' else 'cerebellumPons'
            self._images = [path.join(self._input_params['caps_directory'], 'subjects', self._subjects[i],
                                      self._sessions[i], 'pet/preprocessing/group-' + self._input_params['group_id'],
                                      '%s_%s_task-rest_acq-%s_pet_space-Ixi549Space%s_suvr-%s_mask-brain%s_pet.nii.gz'
                                      % (self._subjects[i], self._sessions[i], self._input_params['image_type'],
                                         pvc, suvr, fwhm))
                            for i in range(len(self._subjects))]

        for image in self._images:
            if not path.exists(image):
                raise Exception("File %s doesn't exists." % image)

        return self._images


class TsvInput(base.MLInput):

    def __init__(self, input_params):

        super().__init__(input_params)

        import pandas as pd

        self._dataframe = pd.io.parsers.read_csv(input_params['data_tsv'], sep='\t')

        if not input_params['columns']:
            raise Exception("List of columns to use as input can not be empty.")

    def get_x(self):
        self._x = self._dataframe.as_matrix(self._input_params['columns'])
        return self._x

    def get_y(self):
        unique = list(set(self._dataframe["diagnosis"]))
        self._y = np.array([unique.index(x) for x in self._dataframe["diagnosis"]])
        return self._y

    def get_kernel(self, kernel_function=utils.gram_matrix_linear, recompute_if_exists=False):
        """
        Returns: a numpy 2d-array.
        """

        if self._kernel is not None and not recompute_if_exists:
            return self._kernel

        if self._x is None:
            self.get_x()

        cprint("Computing kernel ...")
        self._kernel = kernel_function(self._x)
        cprint("Kernel computed")
        return self._kernel

    @staticmethod
    def get_default_parameters():

        parameters_dict = {'data_tsv': None,
                           'columns': None}

        return parameters_dict



# coding: utf-8

import clinica.pipelines.engine as cpe


class StatisticsVolume(cpe.Pipeline):
    """StatisticsVolume - Volume-based mass-univariate analysis with SPM.

    Returns:
        A clinica pipeline object containing the StatisticsVolume pipeline.
    """
    def check_pipeline_parameters(self):
        """Check pipeline parameters."""
        from clinica.utils.exceptions import ClinicaException

        if 'orig_input_data' not in self.parameters.keys():
            raise KeyError('Missing compulsory orig_input_data key in pipeline parameter.')

        if self.parameters['cluster_threshold'] < 0 or self.parameters['cluster_threshold'] > 1:
            raise ClinicaException("Cluster threshold should be between 0 and 1 "
                                   "(given value: %s)." % self.parameters['cluster_threshold'])

    def check_custom_dependencies(self):
        """Check dependencies that can not be listed in the `info.json` file."""
        pass

    def get_input_fields(self):
        """Specify the list of possible inputs of this pipeline.

        Returns:
            A list of (string) input fields name.
        """

        return ['input_files']

    def get_output_fields(self):
        """Specify the list of possible outputs of this pipeline.

        Returns:
            A list of (string) output fields name.
        """

        return ['spmT_0001',
                'spmT_0002',
                'spm_figures',
                'variance_of_error',
                'resels_per_voxels',
                'mask',
                'regression_coeff',
                'contrast']

    def build_input_node(self):
        """Build and connect an input node to the pipeline."""
        from colorama import Fore
        import nipype.interfaces.utility as nutil
        import nipype.pipeline.engine as npe
        from clinica.utils.exceptions import ClinicaException
        from clinica.utils.inputs import clinica_file_reader
        from clinica.utils.input_files import t1_volume_template_tpm_in_mni
        from clinica.utils.stream import cprint
        from clinica.utils.ux import print_images_to_process, print_begin_image

        gic = '*'
        if self.parameters['group_id_caps'] is not None:
            gic = self.parameters['group_id_caps']

        all_errors = []
        if self.parameters['orig_input_data'] == 'pet-volume':
            self.parameters['measure_label'] = 'fdg'
            information_dict = {
                'pattern': '*_pet_space-Ixi549Space_suvr-pons_mask-brain_fwhm-' + str(self.parameters['full_width_at_half_maximum']) + 'mm_pet.nii*',
                'description': 'pons normalized FDG PET image in MNI space (brain masked)',
                'needed_pipeline': 'pet-volume'
            }
        elif self.parameters['orig_input_data'] == 't1-volume':
            self.parameters['measure_label'] = 'graymatter'
            information_dict = t1_volume_template_tpm_in_mni(gic, 0, True)

        else:
            if not self.parameters['custom_file']:
                raise ClinicaException(Fore.RED + '[Error] You did not specify the --custom_file flag in the command line for the feature type '
                                       + Fore.Blue + self.parameters['measure_label'] + Fore.RED + '! Clinica can\'t '
                                       + 'know what file to use in your analysis ! Type: \n\t' + Fore.BLUE + 'clinica run statistics-volume\n'
                                       + Fore.RED + ' to have help on how to use the command line.' + Fore.RESET)
            # If custom file are grabbed, information of fwhm is irrelevant and should not appear on final filenames
            self.parameters['full_width_at_half_maximum'] = None
            information_dict = {
                'pattern': self.parameters['custom_file'],
                'description': 'custom file provided by user'
            }

        try:
            input_files = clinica_file_reader(self.subjects,
                                              self.sessions,
                                              self.caps_directory,
                                              information_dict)
        except ClinicaException as e:
            all_errors.append(e)

        if len(all_errors) > 0:
            error_message = 'Clinica faced errors while trying to read files in your CAPS directories.\n'
            for msg in all_errors:
                error_message += str(msg)
            raise ClinicaException(error_message)

        read_parameters_node = npe.Node(name="LoadingCLIArguments",
                                        interface=nutil.IdentityInterface(
                                            fields=self.get_input_fields(),
                                            mandatory_inputs=True),
                                        synchronize=True)
        read_parameters_node.inputs.input_files = input_files

        if len(self.subjects):
            print_images_to_process(self.subjects, self.sessions)
            cprint('The pipeline will last a few minutes. Images generated by SPM will popup during the pipeline.')
            print_begin_image('group-' + self.parameters['group_label'])

        self.connect([
            (read_parameters_node,      self.input_node,    [('input_files',    'input_files')])
        ])

    def build_output_node(self):
        """Build and connect an output node to the pipeline."""
        import nipype.pipeline.engine as npe
        import nipype.interfaces.io as nio
        from os.path import join, pardir

        relative_path = join('groups',
                             'group-' + self.parameters['group_label'],
                             'statistics_volume',
                             'group_comparison_measure-' + self.parameters['measure_label'])

        datasink = npe.Node(nio.DataSink(), name='sinker')
        datasink.inputs.base_directory = join(self.caps_directory, relative_path)

        datasink.inputs.parameterization = True
        if self.parameters['full_width_at_half_maximum']:
            datasink.inputs.regexp_substitutions = [
                # t-stat map
                (join(self.caps_directory, relative_path) + r'/spm_results_analysis_./(.*)',
                 join(self.caps_directory, relative_path) + r'/\1'),

                # contrasts
                (join(self.caps_directory, relative_path) + r'/contrasts/(.*)',
                 join(self.caps_directory, relative_path) + r'/\1'),

                # resels per voxels
                (join(self.caps_directory, relative_path) + '/resels_per_voxels/resels_per_voxel.nii',
                 join(self.caps_directory, relative_path) + '/group-' + self.parameters['group_label'] + '_RPV.nii'),

                # mask
                (join(self.caps_directory, relative_path) + '/mask/included_voxel_mask.nii',
                 join(self.caps_directory, relative_path) + '/group-' + self.parameters['group_label'] + '_mask.nii'),

                # variance of error
                (join(self.caps_directory, relative_path) + '/variance_of_error/(.*)',
                 join(self.caps_directory, relative_path) + r'/\1'),

                # tsv file
                (join(self.caps_directory, relative_path) + r'/tsv_file/.*',
                 join(self.caps_directory, relative_path) + '/' + pardir + '/group-' + self.parameters['group_label'] + '_participants.tsv'),

                # report (figures)
                (join(self.caps_directory, relative_path) + r'/figures/(.*)',
                 join(self.caps_directory, relative_path) + r'/\1'),

                # regression coefficient
                (join(self.caps_directory, relative_path) + r'/regression_coeff/(.*).nii',
                 join(self.caps_directory, relative_path) + '/group-' + self.parameters['group_label'] + r'_covariate-\1'
                 + '_measure-' + self.parameters['measure_label'] + '_fwhm-'
                 + str(self.parameters['full_width_at_half_maximum']) + '_regressionCoefficient.nii'),

            ]

        datasink.inputs.tsv_file = self.tsv_file

        self.connect([
            (self.output_node, datasink, [('spmT_0001', 'spm_results_analysis_1')]),
            (self.output_node, datasink, [('spmT_0002', 'spm_results_analysis_2')]),
            (self.output_node, datasink, [('spm_figures', 'figures')]),
            (self.output_node, datasink, [('variance_of_error', 'variance_of_error')]),
            (self.output_node, datasink, [('resels_per_voxels', 'resels_per_voxels')]),
            (self.output_node, datasink, [('mask', 'mask')]),
            (self.output_node, datasink, [('regression_coeff', 'regression_coeff')]),
            (self.output_node, datasink, [('contrasts', 'contrasts')])
        ])

    def build_core_nodes(self):
        """Build and connect the core nodes of the pipeline."""
        import clinica.pipelines.statistics_volume.statistics_volume_utils as utils
        import nipype.interfaces.utility as nutil
        import nipype.pipeline.engine as npe
        from clinica.utils.filemanip import unzip_nii
        from os.path import join, dirname

        # SPM cannot handle zipped files
        unzip_node = npe.Node(nutil.Function(input_names=['in_file'],
                                             output_names=['output_files'],
                                             function=unzip_nii),
                              name='unzip_node')

        # Get indexes of the 2 groups, based on the contrast column of the tsv file
        get_groups = npe.Node(nutil.Function(input_names=['tsv', 'contrast'],
                                             output_names=['idx_group1', 'idx_group2', 'class_names'],
                                             function=utils.get_group_1_and_2),
                              name='get_groups')

        get_groups.inputs.contrast = self.parameters['contrast']
        get_groups.inputs.tsv = self.tsv_file

        # Run SPM nodes are all a copy of a generic SPM script launcher
        run_spm_script_node = npe.Node(nutil.Function(input_names=['m_file'],
                                                      output_names=['spm_mat'],
                                                      function=utils.run_m_script),
                                       name='run_spm_script_node')

        run_spm_model_creation = run_spm_script_node.clone(name='run_spm_model_creation')
        run_spm_model_estimation = run_spm_script_node.clone(name='run_spm_model_estimation')
        run_spm_model_contrast = run_spm_script_node.clone(name='run_spm_model_contrast')
        run_spm_model_result_no_correction = run_spm_script_node.clone(name='run_spm_model_result_no_correction')

        # All the following node are creating the correct (.m) script for the different SPM steps
        # 1. Model creation
        # 2. Model estimation
        # 3. Creation of contrast with covariates
        # 4. Creation of results

        # 1. Model creation
        # We use overwrite option to be sure this node is always run so that it can delete the output dir if it
        # already exists (this may cause error in output files otherwise)
        model_creation = npe.Node(nutil.Function(input_names=['tsv',
                                                              'contrast',
                                                              'idx_group1',
                                                              'idx_group2',
                                                              'file_list',
                                                              'template_file'],
                                                 output_names=['script_file', 'covariates'],
                                                 function=utils.model_creation),
                                  name='model_creation',
                                  overwrite=True)
        model_creation.inputs.tsv = self.tsv_file
        model_creation.inputs.contrast = self.parameters['contrast']
        model_creation.inputs.template_file = join(dirname(__file__), 'template_model_creation.m')

        # 2. Model estimation
        model_estimation = npe.Node(nutil.Function(input_names=['mat_file', 'template_file'],
                                                   output_names=['script_file'],
                                                   function=utils.estimate),
                                    name='model_estimation')
        model_estimation.inputs.template_file = join(dirname(__file__), 'template_model_estimation.m')

        # 3. Contrast
        model_contrast = npe.Node(nutil.Function(input_names=['mat_file',
                                                              'template_file',
                                                              'covariates',
                                                              'class_names'],
                                                 output_names=['script_file'],
                                                 function=utils.contrast),
                                  name='model_contrast')
        model_contrast.inputs.template_file = join(dirname(__file__), 'template_model_contrast.m')

        # 4. Results
        model_result_no_correction = npe.Node(nutil.Function(input_names=['mat_file', 'template_file',  'method', 'threshold'],
                                                             output_names=['script_file'],
                                                             function=utils.results),
                                              name='model_result_no_correction')
        model_result_no_correction.inputs.template_file = join(dirname(__file__), 'template_model_results.m')

        model_result_no_correction.inputs.method = 'none'
        model_result_no_correction.inputs.threshold = self.parameters['cluster_threshold']

        # Print result to txt file if spm

        # Export results to output node
        read_output_node = npe.Node(nutil.Function(
            input_names=['spm_mat', 'class_names', 'covariates', 'group_label', 'fwhm', 'measure'],
            output_names=['spmT_0001', 'spmT_0002', 'spm_figures', 'variance_of_error',
                          'resels_per_voxels', 'mask', 'regression_coeff', 'contrasts'],
            function=utils.read_output),
            name='read_output_node')
        read_output_node.inputs.group_label = self.parameters['group_label']
        read_output_node.inputs.fwhm = self.parameters['full_width_at_half_maximum']
        read_output_node.inputs.measure = self.parameters['measure_label']

        # Connection
        # ==========
        self.connect([
            (self.input_node, unzip_node, [('input_files', 'in_file')]),
            (unzip_node, model_creation, [('output_files', 'file_list')]),
            (get_groups, model_creation, [('idx_group1', 'idx_group1')]),
            (get_groups, model_creation, [('idx_group2', 'idx_group2')]),

            (model_creation, run_spm_model_creation, [('script_file', 'm_file')]),
            (run_spm_model_creation, model_estimation, [('spm_mat', 'mat_file')]),
            (model_estimation, run_spm_model_estimation, [('script_file', 'm_file')]),

            (get_groups, model_contrast, [('class_names', 'class_names')]),
            (run_spm_model_estimation, model_contrast, [('spm_mat', 'mat_file')]),
            (model_creation, model_contrast, [('covariates', 'covariates')]),

            (model_contrast, run_spm_model_contrast, [('script_file', 'm_file')]),

            (run_spm_model_contrast, model_result_no_correction, [('spm_mat', 'mat_file')]),

            (model_result_no_correction, run_spm_model_result_no_correction, [('script_file', 'm_file')]),

            (run_spm_model_result_no_correction, read_output_node, [('spm_mat', 'spm_mat')]),

            (get_groups, read_output_node, [('class_names', 'class_names')]),
            (model_creation, read_output_node, [('covariates', 'covariates')]),

            (read_output_node, self.output_node, [('spmT_0001', 'spmT_0001')]),
            (read_output_node, self.output_node, [('spmT_0002', 'spmT_0002')]),
            (read_output_node, self.output_node, [('spm_figures', 'spm_figures')]),
            (read_output_node, self.output_node, [('variance_of_error', 'variance_of_error')]),
            (read_output_node, self.output_node, [('resels_per_voxels', 'resels_per_voxels')]),
            (read_output_node, self.output_node, [('mask', 'mask')]),
            (read_output_node, self.output_node, [('regression_coeff', 'regression_coeff')]),
            (read_output_node, self.output_node, [('contrasts', 'contrasts')]),
        ])



# coding: utf8


class Manage_conflicts():

    def __init__(self, path_choices):

        self.dic = self.parse_choices(path_choices)

    def parse_choices(self, path_choices):
        f = open(path_choices, 'r')
        dic = dict()
        for line in f.readlines():
            line = line.split('-')
            dic[line[0]] = line[1]

        return dic

    def make_decision(self, list_med_names):
        list_med_names.sort()
        return self.dic[str(list_med_names)]

    def __str__(self):
        s = ''
        for key in self.dic:
            s += str(key) + ' -> ' + str(self.dic[key])

        return s



# coding: utf8


def print_statistics(summary_file, num_subjs, ses_aval, mmt):
    """
    Print to a given input file statistics about missing files and modalities
    in a dataset.  This metod takes in input a MissingModsTracker object (mmt)
    that contains the number of missing modalities for each session and the
    number of missing sessions for each subject.

    Args:
        summary_file: path of the output file where write.
        num_subjs: number of subjects.
        ses_aval: list of sessions available.
        mmt: object MissingModsTracker
    """
    missing_list = mmt.get_missing_list()
    summary_file.write('**********************************************\n')
    summary_file.write('Number of subjects converted: ' + str(num_subjs) + '\n')
    summary_file.write('Sessions available: '+ses_aval[0] + '\n')

    for ses in ses_aval:
        ses_miss = missing_list[ses]['session']
        ses_found = num_subjs - ses_miss
        perc_ses_found = ses_found*100/num_subjs
        summary_file.write('Number of sessions '+ses+" found: "+str(ses_found)+' ('+str(perc_ses_found)+'%)\n')

    summary_file.write('**********************************************\n\n')
    summary_file.write('Number of missing modalities for each session:\n')

    for ses in ses_aval:
        summary_file.write('\n' + ses + '\n')
        for mod in missing_list[ses]:
            if mod != 'session':
                num_miss_mod = missing_list[ses][mod]
                percentage_missing = round((num_miss_mod*100/float(num_subjs - missing_list[ses]['session'])), 2)
                summary_file.write(mod+': ' + str(num_miss_mod) + ' ('+str(percentage_missing) + '%) \n')


def has_one_index(index_list):
    if len(index_list) == 1:
        return index_list[0]
    if len(index_list) == 0:
        return -1
    if len(index_list) > 1:
        raise('Multiple indexes found')


class MissingModsTracker:
    """
    Class used for tracking the number of missing modalities in a database
    """
    def __init__(self, ses, mod_list=False):
        self.missing = {}
        self.ses = ses
        if mod_list:
            for s in ses:
                self.missing.update({s: {'session': 0}})
                for mod in mod_list:
                    self.missing[s].update({mod: 0})
        else:
            for s in ses:
                self.missing.update({
                    s: {'session': 0,
                        'dwi': 0,
                        'func': 0,
                        'fieldmap': 0,
                        'flair': 0,
                        't1w': 0}
                    })

    def add_missing_mod(self, ses, mod):
        """
        Increase the number of missing files for the input modality.

        Args:
            ses: name of the session
            mod: modality missing
        """
        self.missing[ses][mod] += 1

    def increase_missing_ses(self, ses):
        self.missing[ses]['session'] += 1

    def get_missing_list(self):
        """
        Return the hash map mods_missing.

        Returns:
             The hash map containing the list of missing files.

        """
        return self.missing



# coding: utf8

"""
This module contains utilities for longitudinal pipelines.

See CAPS specifications for details about long ID.
"""

from clinica.utils.filemanip import read_participant_tsv


def get_unique_subjects(in_subject_list, in_session_list):
    """Get unique participant IDs with their sessions.

    This function generates a list of unique participant IDs from `in_subject_list` with their sessions.
    Args:
        in_subject_list (list[str]): list of participant IDs
        in_session_list (list[str]): list of session IDs

    Returns:
        out_unique_subject_list (list[str]): list of participant IDs, where each participant appears only once
        out_per_subject_session_list (list[list[str]]): list of list
            (list2) of session_id associated to any single participant

    Example:
        >>> from clinica.utils.participant import get_unique_subjects
        >>> get_unique_subjects(['sub-CLNC01', 'sub-CLNC01', 'sub-CLNC02'], ['ses-M00', 'ses-M18', 'ses-M00'])
        (['sub-CLNC01', 'sub-CLNC02'], [['ses-M00', 'ses-M18'], ['ses-M00']])
    """
    import numpy as np

    subject_array = np.array(in_subject_list)
    session_array = np.array(in_session_list)

    # The second returned element indicates for each participant_id the
    # element they correspond to in the 'unique' list. We will use this
    # to link each session_id in the repeated list of session_id to
    # their corresponding unique participant_id

    unique_subject_array, out_inverse_positions = np.unique(
        subject_array, return_inverse=True)
    out_unique_subject_list = unique_subject_array.tolist()

    subject_number = len(out_unique_subject_list)
    out_per_subject_session_list = [
        session_array[
            out_inverse_positions == subject_index
            ].tolist() for subject_index in range(subject_number)]

    assert len(out_unique_subject_list) == len(out_per_subject_session_list),\
        'Problem while getting unique subjects and sessions lists'

    return out_unique_subject_list, out_per_subject_session_list


def unique_subjects_sessions_to_subjects_sessions(unique_subject_list, per_subject_session_list):
    """Do reverse operation of get_unique_subjects function.

    Example:
        >>> from clinica.utils.participant import unique_subjects_sessions_to_subjects_sessions
        >>> unique_subjects_sessions_to_subjects_sessions(['sub-01', 'sub-02'], [['ses-M00', 'ses-M18'], ['ses-M00']])
        (['sub-CLNC01', 'sub-01', 'sub-02'], ['ses-M00', 'ses-M18', 'ses-M00'])

    """
    list_participants = []
    list_sessions = []
    for idx, participant_id in enumerate(unique_subject_list):
        for session_id in per_subject_session_list[idx]:
            list_participants.append(participant_id)
            list_sessions.append(session_id)

    return list_participants, list_sessions


def get_subject_session_list(input_dir, ss_file=None, is_bids_dir=True, use_session_tsv=False, tsv_dir=None):
    """Parse a BIDS or CAPS directory to get the subjects and sessions.

    This function lists all the subjects and sessions based on the content of
    the BIDS or CAPS directory or (if specified) on the provided
    subject-sessions TSV file.

    Args:
        input_dir: A BIDS or CAPS directory path.
        ss_file: A subjects-sessions file (.tsv format).
        is_bids_dir: Indicates if input_dir is a BIDS or CAPS directory
        use_session_tsv (boolean): Specify if the list uses the sessions listed in the sessions.tsv files
        tsv_dir (str): if TSV file does not exist, it will be created in output_dir. If
            not specified, output_dir will be in <tmp> folder

    Returns:
        subjects: A subjects list.
        sessions: A sessions list.

    Notes:
        This is a generic method based on folder names. If your <BIDS> dataset contains e.g.:
        - sub-CLNC01/ses-M00/anat/sub-CLNC01_ses-M00_T1w.nii
        - sub-CLNC02/ses-M00/dwi/sub-CLNC02_ses-M00_dwi.{bval|bvec|json|nii}
        - sub-CLNC02/ses-M00/anat/sub-CLNC02_ses-M00_T1w.nii
        get_subject_session_list(<BIDS>, None, True) will return
        ['ses-M00', 'ses-M00'], ['sub-CLNC01', 'sub-CLNC02'].

        However, if your pipeline needs both T1w and DWI files, you will need to check
        with e.g. clinica_file_reader_function.
    """
    import os
    import tempfile
    from time import time, strftime, localtime
    import clinica.iotools.utils.data_handling as cdh

    if not ss_file:
        if tsv_dir:
            output_dir = tsv_dir
        else:
            output_dir = tempfile.mkdtemp()
        timestamp = strftime('%Y%m%d_%H%M%S', localtime(time()))
        tsv_file = 'subjects_sessions_list_%s.tsv' % timestamp
        ss_file = os.path.join(output_dir, tsv_file)

        cdh.create_subs_sess_list(
            input_dir=input_dir,
            output_dir=output_dir,
            file_name=tsv_file,
            is_bids_dir=is_bids_dir,
            use_session_tsv=use_session_tsv)

    participant_ids, session_ids = read_participant_tsv(ss_file)
    return session_ids, participant_ids


def have_same_subjects(tsv_file_1, tsv_file_2):
    """Return True if `tsv_file_1` and `tsv_file_2` have the same subjects, False otherwise."""
    import pandas as pd

    tsv_df_1 = pd.io.parsers.read_csv(tsv_file_1, sep='\t')
    tsv_df_2 = pd.io.parsers.read_csv(tsv_file_2, sep='\t')
    image_ids_1 = [
        p_id + '_' + s_id
        for (p_id, s_id) in zip(list(tsv_df_1.participant_id), list(tsv_df_1.session_id))
    ]
    image_ids_2 = [
        p_id + '_' + s_id
        for (p_id, s_id) in zip(list(tsv_df_2.participant_id), list(tsv_df_2.session_id))
    ]
    diff_image_ids = list(set(image_ids_1) - set(image_ids_2))

    if len(diff_image_ids) == 0:
        return True
    else:
        return False



# coding: utf8

import clinica.engine as ce


class T1FreeSurferLongitudinalCLI(ce.CmdParser):

    def define_name(self):
        """Define the sub-command name to run this pipeline."""
        self._name = 't1-freesurfer-longitudinal'

    def define_description(self):
        """Define a description of this pipeline."""
        self._description = ('Longitudinal pre-processing of T1w images with FreeSurfer:\n'
                             'http://clinica.run/doc/Pipelines/T1_FreeSurfer_Longitudinal/')

    def define_options(self):
        from clinica.engine.cmdparser import PIPELINE_CATEGORIES
        # Clinica compulsory arguments (e.g. BIDS, CAPS, group_id)
        clinica_comp = self._args.add_argument_group(PIPELINE_CATEGORIES['CLINICA_COMPULSORY'])
        clinica_comp.add_argument("caps_directory",
                                  help='Path to the CAPS directory.')

        # Clinica standard arguments (e.g. --n_procs)
        self.add_clinica_standard_arguments(add_overwrite_flag=True)

    def run_command(self, args):
        """Run the pipeline with defined args."""
        import os
        import datetime
        from colorama import Fore
        from clinica.utils.stream import cprint
        from clinica.utils.longitudinal import get_participants_long_id
        from clinica.utils.participant import get_subject_session_list
        from .t1_freesurfer_template_cli import T1FreeSurferTemplateCLI
        from .t1_freesurfer_longitudinal_correction_cli import T1FreeSurferLongitudinalCorrectionCLI
        from .longitudinal_utils import save_part_sess_long_ids_to_tsv

        cprint(
            'The t1-freesurfer-longitudinal pipeline is divided into 2 parts:\n'
            '\t%st1-freesurfer-unbiased-template pipeline%s: Creation of unbiased template\n'
            '\t%st1-freesurfer-longitudinal-correction pipeline%s: Longitudinal correction\n'
            % (Fore.BLUE, Fore.RESET, Fore.BLUE, Fore.RESET)
        )

        if not self.absolute_path(args.subjects_sessions_tsv):
            l_sess, l_part = get_subject_session_list(self.absolute_path(args.caps_directory), None, False, False)
            l_long = get_participants_long_id(l_part, l_sess)
            now = datetime.datetime.now().strftime('%H%M%S')
            args.subjects_sessions_tsv = now + '_participants.tsv'
            save_part_sess_long_ids_to_tsv(l_part, l_sess, l_long, os.getcwd(), args.subjects_sessions_tsv)

        cprint('%s\nPart 1/2: Running t1-freesurfer-unbiased-template pipeline%s' % (Fore.BLUE, Fore.RESET))
        unbiased_template_cli = T1FreeSurferTemplateCLI()
        unbiased_template_cli.run_command(args)

        cprint('%s\nPart 2/2 Running t1-freesurfer-longitudinal-correction pipeline%s' % (Fore.BLUE, Fore.RESET))
        longitudinal_correction_cli = T1FreeSurferLongitudinalCorrectionCLI()
        longitudinal_correction_cli.run_command(args)



# coding: utf8


def get_substitutions_datasink(bids_file):

    substitutions_ls = [  # registration
            (bids_file + '_T1w_corrected.nii.gz',
                bids_file + '_desc-BiasCorrected_T1w.nii.gz'),
            (bids_file + 'Warped_cropped_intensity_norm.nii.gz',
                bids_file + '_space-MNI152NLin2009cSym_res-1x1x1_intensity_norm_T1w.nii.gz'),
            (bids_file + 'Warped_cropped.nii.gz',
                bids_file + '_space-MNI152NLin2009cSym_desc-Crop_res-1x1x1_T1w.nii.gz'),
            (bids_file + '0GenericAffine.mat',
                bids_file + '_space-MNI152NLin2009cSym_res-1x1x1_affine.mat'),
            (bids_file + 'Warped.nii.gz',
                bids_file + '_space-MNI152NLin2009cSym_res-1x1x1_T1w.nii.gz')
            ]
    return bids_file, substitutions_ls


# Function used by the nipype interface.
# It crops an image based on the reference.
def crop_nifti(input_img, ref_crop):
    """Crop input image based on the reference. It uses nilearn
    `resample_to_img` function.
    Args:
       input_img (str): image to be processed
       ref_img (str): template used to crop the image
    Returns:
       output_img (nifty image): crop image on disk.
       crop_template: (nifty image): output template on disk.
    """

    import nibabel as nib
    import os
    import numpy as np
    from nilearn.image import resample_img, crop_img, resample_to_img
    from nibabel.spatialimages import SpatialImage

    basedir = os.getcwd()
    # crop_ref = crop_img(ref_img, rtol=0.5)
    # crop_ref.to_filename(os.path.join(basedir, os.path.basename(input_img).split('.nii')[0] + '_cropped_template.nii.gz'))
    # crop_template = os.path.join(basedir, os.path.basename(input_img).split('.nii')[0] + '_cropped_template.nii.gz')

    # resample the individual MRI into the cropped template image
    crop_img = resample_to_img(input_img, ref_crop, force_resample=True)
    crop_img.to_filename(os.path.join(basedir, os.path.basename(input_img).split('.nii')[0] + '_cropped.nii.gz'))

    output_img = os.path.join(basedir, os.path.basename(input_img).split('.nii')[0] + '_cropped.nii.gz')
    crop_template = ref_crop

    return output_img, crop_template


def print_end_pipeline(t1w, final_file):
    """
    Display end message for <subject_id> when <final_file> is connected.
    """
    from clinica.utils.ux import print_end_image
    from clinica.utils.filemanip import get_subject_id
    print_end_image(get_subject_id(t1w))



# coding: utf-8

import clinica.pipelines.engine as cpe


class StatisticsVolumeCorrection(cpe.Pipeline):
    """StatisticsVolumeCorrection - Statistical correction of Statistical correction of StatisticsVolume pipeline.

    Returns:
        A clinica pipeline object containing the StatisticsVolumeCorrection pipeline.
    """

    def check_custom_dependencies(self):
        """Check dependencies that can not be listed in the `info.json` file.
        """
        pass

    def get_input_fields(self):
        """Specify the list of possible inputs of this pipeline.

        Returns:
            A list of (string) input fields name.
        """
        return ['t_map']

    def get_output_fields(self):
        """Specify the list of possible outputs of this pipeline.

        Returns:
            A list of (string) output fields name.
        """
        return []

    def build_input_node(self):
        """Build and connect an input node to the pipeline."""
        import nipype.interfaces.utility as nutil
        import nipype.pipeline.engine as npe
        from clinica.utils.inputs import clinica_group_reader

        t_map = clinica_group_reader(self.caps_directory, {'pattern': self.parameters['t_map'] + '*',
                                                           'description': 'statistics t map',
                                                           'needed_pipeline': 'statistics-volume'})

        read_parameters_node = npe.Node(name="LoadingCLIArguments",
                                        interface=nutil.IdentityInterface(
                                            fields=self.get_input_fields(),
                                            mandatory_inputs=True))
        read_parameters_node.inputs.t_map = t_map

        self.connect([
            (read_parameters_node,      self.input_node,    [('t_map', 't_map')])
        ])

    def build_output_node(self):
        """Build and connect an output node to the pipeline."""
        pass

    def build_core_nodes(self):
        """Build and connect the core nodes of the pipeline."""
        import clinica.pipelines.statistics_volume_correction.statistics_volume_correction_utils as utils
        from clinica.utils.inputs import fetch_file, RemoteFileStructure
        import nipype.interfaces.utility as nutil
        import nipype.pipeline.engine as npe
        from os.path import join, abspath, pardir, dirname, exists
        import numpy as np

        peak_correction_FWE = npe.Node(name='peak_correction_FWE',
                                       interface=nutil.Function(
                                           input_names=['t_map', 't_threshold'],
                                           output_names=['output'],
                                           function=utils.peak_correction))
        peak_correction_FWE.inputs.t_threshold = self.parameters['FWEp']

        peak_correction_FDR = peak_correction_FWE.clone(name='peak_correction_FDR')
        peak_correction_FDR.inputs.t_threshold = self.parameters['FDRp']

        cluster_correction_FWE = npe.Node(name='cluster_correction_FWE',
                                          interface=nutil.Function(
                                              input_names=['t_map', 't_thresh', 'c_thresh'],
                                              output_names=['output'],
                                              function=utils.cluster_correction))
        cluster_correction_FWE.inputs.t_thresh = self.parameters['height_threshold']
        cluster_correction_FWE.inputs.c_thresh = self.parameters['FWEc']

        cluster_correction_FDR = cluster_correction_FWE.clone(name='cluster_correction_FDR')
        cluster_correction_FDR.inputs.t_thresh = self.parameters['height_threshold']
        cluster_correction_FDR.inputs.c_thresh = self.parameters['FDRc']

        produce_fig_FWE_peak_correction = npe.Node(name='produce_figure_FWE_peak_correction',
                                                   interface=nutil.Function(
                                                       input_names=['nii_file', 'template', 'type_of_correction', 't_thresh', 'c_thresh', 'n_cuts'],
                                                       output_names=['figs'],
                                                       function=utils.produce_figures))
        produce_fig_FWE_peak_correction.inputs.n_cuts = self.parameters['n_cuts']

        root = dirname(abspath(join(abspath(__file__), pardir, pardir)))
        path_to_mask = join(root, 'resources', 'masks')
        url_aramis = 'https://aramislab.paris.inria.fr/files/data/img_t1_linear/'
        FILE1 = RemoteFileStructure(
                filename='mni_icbm152_t1_tal_nlin_sym_09a.nii.gz',
                url=url_aramis,
                checksum='3b244ee7e287319d36a25263744c468ef0ab2fe5a94b15a2138844db73b49adf'
                )

        produce_fig_FWE_peak_correction.inputs.template = join(path_to_mask, FILE1.filename)
        if not(exists(produce_fig_FWE_peak_correction.inputs.template)):
            try:
                fetch_file(FILE1, path_to_mask)
            except IOError as err:
                cprint('Unable to download required template (mni_icbm152) for processing:', err)

        produce_fig_FDR_peak_correction = produce_fig_FWE_peak_correction.clone(name='produce_figure_FDR_peak_correction')
        produce_fig_FWE_cluster_correction = produce_fig_FWE_peak_correction.clone(name='produce_figure_FWE_cluster_correction')
        produce_fig_FDR_cluster_correction = produce_fig_FWE_peak_correction.clone(name='produce_figure_FDR_cluster_correction')

        produce_fig_FWE_peak_correction.inputs.type_of_correction = 'FWE'
        produce_fig_FDR_peak_correction.inputs.type_of_correction = 'FDR'
        produce_fig_FWE_cluster_correction.inputs.type_of_correction = 'FWE'
        produce_fig_FDR_cluster_correction.inputs.type_of_correction = 'FDR'

        produce_fig_FWE_peak_correction.inputs.t_thresh = self.parameters['FWEp']
        produce_fig_FDR_peak_correction.inputs.t_thresh = self.parameters['FDRp']
        produce_fig_FWE_cluster_correction.inputs.t_thresh = self.parameters['height_threshold']
        produce_fig_FDR_cluster_correction.inputs.t_thresh = self.parameters['height_threshold']

        produce_fig_FWE_peak_correction.inputs.c_thresh = np.nan
        produce_fig_FDR_peak_correction.inputs.c_thresh = np.nan
        produce_fig_FWE_cluster_correction.inputs.c_thresh = self.parameters['FWEc']
        produce_fig_FDR_cluster_correction.inputs.c_thresh = self.parameters['FDRc']

        save_fig_peak_correction_FWE = npe.Node(name='save_figure_peak_correction_FWE',
                                                interface=nutil.Function(
                                                    input_names=['t_map', 'figs', 'name'],
                                                    output_names=[],
                                                    function=utils.generate_output))
        save_fig_peak_correction_FWE.inputs.name = 'FWEp'

        save_fig_peak_correction_FDR = save_fig_peak_correction_FWE.clone(name='save_fig_peak_correction_FDR')
        save_fig_peak_correction_FDR.inputs.name = 'FDRp'

        save_fig_cluster_correction_FWE = save_fig_peak_correction_FWE.clone(name='save_fig_cluster_correction_FWE')
        save_fig_cluster_correction_FWE.inputs.name = 'FWEc'

        save_fig_cluster_correction_FDR = save_fig_peak_correction_FWE.clone(name='save_fig_cluster_correction_FDR')
        save_fig_cluster_correction_FDR.inputs.name = 'FDRc'

        # Connection
        # ==========
        self.connect([
            (self.input_node, peak_correction_FWE, [('t_map', 't_map')]),
            (self.input_node, peak_correction_FDR, [('t_map', 't_map')]),
            (self.input_node, cluster_correction_FWE, [('t_map', 't_map')]),
            (self.input_node, cluster_correction_FDR, [('t_map', 't_map')]),

            (peak_correction_FWE, produce_fig_FWE_peak_correction, [('output', 'nii_file')]),
            (peak_correction_FDR, produce_fig_FDR_peak_correction, [('output', 'nii_file')]),
            (cluster_correction_FWE, produce_fig_FWE_cluster_correction, [('output', 'nii_file')]),
            (cluster_correction_FDR, produce_fig_FDR_cluster_correction, [('output', 'nii_file')]),

            (produce_fig_FWE_peak_correction, save_fig_peak_correction_FWE, [('figs', 'figs')]),
            (produce_fig_FDR_peak_correction, save_fig_peak_correction_FDR, [('figs', 'figs')]),
            (produce_fig_FWE_cluster_correction, save_fig_cluster_correction_FWE, [('figs', 'figs')]),
            (produce_fig_FDR_cluster_correction, save_fig_cluster_correction_FDR, [('figs', 'figs')]),

            (self.input_node, save_fig_peak_correction_FWE, [('t_map', 't_map')]),
            (self.input_node, save_fig_peak_correction_FDR, [('t_map', 't_map')]),
            (self.input_node, save_fig_cluster_correction_FWE, [('t_map', 't_map')]),
            (self.input_node, save_fig_cluster_correction_FDR, [('t_map', 't_map')])
        ])
